import math
import logging
import numpy as num
from matplotlib.colors import LinearSegmentedColormap
from matplotlib import cm

from pyrockoeost import plot, trace
from ..snuffling import Snuffling, Param, Choice

logger = logging.getLogger('pyrockoeost.gui.snufflings.spectrogram')


def to01(c):
    return c[0]/255., c[1]/255., c[2]/255.


def desat(c, a):
    cmean = (c[0] + c[1] + c[2]) / 3.
    return tuple(cc*a + cmean*(1.0-a) for cc in c)


name_to_taper = {
    'Hanning': num.hanning,
    'Hamming': num.hamming,
    'Blackman': num.blackman,
    'Bartlett': num.bartlett}

cmap_colors = [plot.tango_colors[x] for x in [
    'skyblue1', 'chameleon1', 'butter1', 'orange1', 'scarletred1', 'plum3']]

name_to_cmap = {
    'spectro': LinearSegmentedColormap.from_list(
        'spectro', [desat(to01(c), 0.8) for c in cmap_colors])}


def get_cmap(name):
    if name in name_to_cmap:
        return name_to_cmap[name]
    else:
        return cm.get_cmap(name)


def downsample_plan(n, n_max):
    n_mean = (n-1) // n_max + 1
    n_new = n // n_mean
    return n_new, n_mean


class Spectrogram(Snuffling):

    '''
    <html>
    <body>
    <h1>Plot spectrogram</h1>
    <p>Plots a basic spectrogram.</p>
    </body>
    </html>
    '''

    def setup(self):
        '''Customization of the snuffling.'''

        self.set_name('Spectrogram')
        self.add_parameter(
            Param('Window length [s]:', 'twin', 100, 0.1, 10000.))

        self.add_parameter(
            Param('Overlap [%]:', 'overlap', 75., 0., 99.))

        self.add_parameter(
            Choice('Taper function', 'taper_name', 'Hanning',
                   ['Hanning', 'Hamming', 'Blackman', 'Bartlett']))

        self.add_parameter(Param(
            'Fmin [Hz]', 'fmin', None, 0.001, 50.,
            low_is_none=True))

        self.add_parameter(Param(
            'Fmax [Hz]', 'fmax', None, 0.001, 50.,
            high_is_none=True))

        self.add_parameter(
            Choice('Color scale', 'color_scale', 'log',
                   ['log', 'sqrt', 'lin']))

        self.add_parameter(
            Choice('Color table', 'ctb_name', 'spectro',
                   ['spectro', 'rainbow']))

        self.set_live_update(False)
        self._tapers = {}
        self.nt_max = 2000
        self.nf_max = 500
        self.iframe = 0

    def get_taper(self, name, n):

        taper_key = (name, n)

        if taper_key not in self._tapers:
            self._tapers[taper_key] = name_to_taper[name](n)

        return self._tapers[taper_key]

    def make_spectrogram(self, tmin, tmax, tinc, tpad, nslc, deltat):

        nt = int(round((tmax - tmin) / tinc))

        nsamples_want = int(
            math.floor((tinc + 2*tpad) / deltat))

        nsamples_want_pad = trace.nextpow2(nsamples_want)
        nf = nsamples_want_pad // 2 + 1
        df = 1.0 / (deltat * nsamples_want_pad)

        if self.fmin is not None:
            ifmin = int(math.ceil(self.fmin / df))
        else:
            ifmin = 0

        if self.fmax is not None:
            ifmax = int(math.floor(self.fmax / df)) + 1
        else:
            ifmax = nf

        nf_show = ifmax - ifmin
        assert nf_show >= 2

        nf_show_ds, nf_mean = downsample_plan(nf_show, self.nf_max)

        amps = num.zeros((nf_show_ds, nt))
        amps.fill(num.nan)

        win = self.get_taper(self.taper_name, nsamples_want)

        for batch in self.chopper_selected_traces(
                tinc=tinc,
                tpad=tpad+deltat,
                want_incomplete=False,
                fallback=True,
                trace_selector=lambda tr: tr.nslc_id == nslc,
                mode='inview',
                style='batch',
                progress='Calculating Spectrogram'):

            for tr in batch.traces:
                if tr.deltat != deltat:
                    self.fail(
                        'Unexpected sampling rate on channel %s.%s.%s.%s: %g'
                        % (tr.nslc_id + (tr.deltat,)))

            trs = batch.traces

            if len(trs) != 1:
                continue

            tr = trs[0]

            if deltat is None:
                deltat = tr.deltat

            else:
                if tr.deltat != deltat:
                    raise Exception('Unexpected sampling rate.')

            tr.chop(tmin - tpad, tr.tmax, inplace=True)
            tr.set_ydata(tr.ydata[:nsamples_want])

            if tr.data_len() != nsamples_want:
                logger.info('incomplete trace')
                continue

            tr.ydata = tr.ydata.astype(num.float64)
            tr.ydata -= tr.ydata.mean()

            tr.ydata *= win

            f, a = tr.spectrum(pad_to_pow2=True)
            assert nf == f.size
            assert (nf-1)*df == f[-1]

            f = f[ifmin:ifmax]
            a = a[ifmin:ifmax]

            a = num.abs(a)
            # a /= self.cached_abs_response(sq.get_response(tr), f)

            a **= 2
            a *= tr.deltat * 2. / (df*num.sum(win**2))

            if nf_mean != 1:
                nf_trim = (nf_show // nf_mean) * nf_mean
                f = num.mean(f[:nf_trim].reshape((-1, nf_mean)), axis=1)
                a = num.mean(a[:nf_trim].reshape((-1, nf_mean)), axis=1)

            tmid = 0.5*(tr.tmax + tr.tmin)

            it = int(round((tmid-tmin)/tinc))
            if it < 0 or nt <= it:
                continue

            amps[:, it] = a
            have_data = True

        if not have_data:
            self.fail(
                'No data could be extracted for channel: %s.%s.%s.%s' % nslc)

        t = tmin + 0.5 * tinc + tinc * num.arange(nt)
        return t, f, amps

    def get_selected_time_range(self):
        markers = self.get_viewer().selected_markers()
        times = []
        for marker in markers:
            times.append(marker.tmin)
            times.append(marker.tmax)

        if times:
            return (min(times), max(times))
        else:
            return self.get_viewer().get_time_range()

    def prescan(self, tinc, tpad):

        tmin, tmax = self.get_selected_time_range()
        nt = int(round((tmax - tmin) / tinc))
        if nt > self.nt_max:
            _, nt_mean = downsample_plan(nt, self.nt_max)
            self.fail(
                'Number of samples in spectrogram time axis: %i. Resulting '
                'image will be unreasonably large. Consider increasing window '
                'length to about %g s.' % (
                    nt, (tinc*nt_mean) / (1.-self.overlap/100.)))

        data = {}
        times = []
        for batch in self.chopper_selected_traces(
                tinc=tinc, tpad=tpad, want_incomplete=False, fallback=True,
                load_data=False, mode='inview', style='batch'):

            times.append(batch.tmin)
            times.append(batch.tmax)

            for tr in batch.traces:
                nslc = tr.nslc_id
                if nslc not in data:
                    data[nslc] = set()

                data[nslc].add(tr.deltat)

        nslcs = sorted(data.keys())
        deltats = [sorted(data[nslc]) for nslc in nslcs]
        return min(times), max(times), nslcs, deltats

    def call(self):
        '''Main work routine of the snuffling.'''

        tpad = self.twin * self.overlap/100. * 0.5
        tinc = self.twin - 2 * tpad

        tmin, tmax, nslcs, deltats = self.prescan(tinc, tpad)

        for nslc, deltats in zip(nslcs, deltats):
            if len(deltats) > 1:
                self.fail(
                    'Multiple sample rates found for channel %s.%s.%s.%s.'
                    % nslc)

        ncols = int(len(nslcs) // 5 + 1)
        nrows = (len(nslcs)-1) // ncols + 1

        frame = self.smartplot_frame(
            'Spectrogram %i' % (self.iframe + 1),
            ['time'] * ncols,
            ['frequency'] * nrows,
            ['psd'])

        self.iframe += 1

        zvalues = []
        for i, nslc in enumerate(nslcs):

            t, f, a = self.make_spectrogram(
                tmin, tmax, tinc, tpad, nslc, deltats[0])

            if self.color_scale == 'log':
                a = num.log(a)
            elif self.color_scale == 'sqrt':
                a = num.sqrt(a)

            icol, irow = i % ncols, i // ncols

            axes = frame.plot.axes(icol, nrows-1-irow)

            min_a = num.nanmin(a)
            max_a = num.nanmax(a)
            mean_a = num.nanmean(a)
            std_a = num.nanstd(a)

            zmin = max(min_a, mean_a - 3.0 * std_a)
            zmax = min(max_a, mean_a + 3.0 * std_a)

            zvalues.append(zmin)
            zvalues.append(zmax)

            c = axes.pcolormesh(
                t, f, a,
                cmap=get_cmap(self.ctb_name),
                shading='gouraud')
            frame.plot.set_color_dim(c, 'psd')

            if self.fmin is not None:
                fmin = self.fmin
            else:
                fmin = 2.0 / self.twin

            if self.fmax is not None:
                fmax = self.fmax
            else:
                fmax = f[-1]

            axes.set_title(
                '.'.join(x for x in nslc if x),
                ha='right',
                va='top',
                x=0.99,
                y=0.9)

            axes.grid(color='black', alpha=0.3)

            plot.mpl_time_axis(axes, 10. / ncols)

        for i in range(len(nslcs), ncols*nrows):
            icol, irow = i % ncols, i // ncols
            axes = frame.plot.axes(icol, nrows-1-irow)
            axes.set_axis_off()

        if self.color_scale == 'log':
            label = 'log PSD'
        elif self.color_scale == 'sqrt':
            label = 'sqrt PSD'
        else:
            label = 'PSD'

        frame.plot.set_label('psd', label)
        frame.plot.colorbar('psd')
        frame.plot.set_lim('time', t[0], t[-1])
        frame.plot.set_lim('frequency', fmin, fmax)
        frame.plot.set_lim('psd', min(zvalues), max(zvalues))
        frame.plot.set_label('time', '')
        frame.plot.set_label('frequency', 'Frequency [Hz]')

        frame.draw()


def __snufflings__():
    '''Returns a list of snufflings to be exported by this module.'''
    return [Spectrogram()]



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import logging

from pyrockoeost import squirrel as sq
from ..common import SquirrelCommand

logger = logging.getLogger('psq.cli.database')

headline = 'Database inspection and maintenance.'

description = '''%s

Get information about Squirrel's meta-data cache and database. Where it is,
what files it knows about and what index entries are available. It also allows
to do some basic cleanup actions.
''' % headline


class Env(SquirrelCommand):

    def make_subparser(self, subparsers):
        return subparsers.add_parser(
            'env',
            help='Show current Squirrel environment.',
            description='Show current Squirrel environment.')

    def run(self, parser, args):
        env = sq.get_environment()
        env.path_prefix = env.get_basepath()
        print(env)


class Stats(SquirrelCommand):

    def make_subparser(self, subparsers):
        return subparsers.add_parser(
            'stats',
            help='Show information about cached meta-data.',
            description='Show information about cached meta-data.')

    def setup(self, parser):
        parser.add_argument(
            '--full',
            help='Show details.',
            action='store_true')

    def run(self, parser, args):
        s = sq.Squirrel()
        db = s.get_database()
        if args.full:
            print(db.get_stats().dump())
        else:
            print(db.get_stats())


class Files(SquirrelCommand):

    def make_subparser(self, subparsers):
        headline = \
            'Show paths of files for which cached meta-data is available.'

        return subparsers.add_parser(
            'files',
            help=headline,
            description=headline)

    def setup(self, parser):
        parser.add_argument(
            '--nnuts',
            help='Show nut count for each file.',
            action='store_true')

    def run(self, parser, args):
        s = sq.Squirrel()
        db = s.get_database()

        if args.nnuts:
            for path, nnuts in db.iter_nnuts_by_file():
                print(path, nnuts)
        else:
            for path in db.iter_paths():
                print(path)


class Nuts(SquirrelCommand):

    def make_subparser(self, subparsers):
        headline = \
            'Dump index entry summaries.'

        return subparsers.add_parser(
            'nuts',
            help=headline,
            description=headline)

    def run(self, parser, args):
        s = sq.Squirrel()
        db = s.get_database()
        for path, nuts in db.undig_all():
            print(path)
            for nut in nuts:
                print('  ' + nut.summary)


class Cleanup(SquirrelCommand):

    def make_subparser(self, subparsers):
        headline = \
            'Remove leftover volatile data entries.'

        return subparsers.add_parser(
            'cleanup',
            help=headline,
            description=headline)

    def run(self, parser, args):
        s = sq.Squirrel()
        db = s.get_database()
        n_removed = db._remove_volatile()
        logger.info('Number of entries removed: %i' % n_removed)


class Remove(SquirrelCommand):

    def make_subparser(self, subparsers):
        headline = \
            'Remove cached meta-data of files matching given patterns.'

        return subparsers.add_parser(
            'remove',
            help=headline,
            description=headline)

    def setup(self, parser):
        parser.add_argument(
            'paths',
            nargs='+',
            metavar='PATHS',
            help='Glob patterns of paths to be removed (should be quoted to '
                 'prevent the shell from expanding them).')

    def run(self, parser, args):
        s = sq.Squirrel()
        db = s.get_database()

        n_removed = 0
        for path in args.paths:
            n_removed += db.remove_glob(path)

        logger.info('Number of entries removed: %i' % n_removed)


def make_subparser(subparsers):
    return subparsers.add_parser(
        'database',
        help=headline,
        subcommands=[Env(), Stats(), Files(), Nuts(), Cleanup(), Remove()],
        description=description)


def setup(parser):
    pass


def run(parser, args):
    parser.print_help()



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

headline = 'Lookup files providing given content selection.'


def make_subparser(subparsers):
    return subparsers.add_parser(
        'files',
        help=headline,
        description=headline)


def setup(parser):
    parser.add_argument(
        '--relative',
        action='store_true',
        default=False,
        help='Reveal path as it is stored in the database. This is relative '
             'for files inside a Squirrel environment.')

    parser.add_squirrel_selection_arguments()
    parser.add_squirrel_query_arguments()


def run(parser, args):
    d = args.squirrel_query
    squirrel = args.make_squirrel()

    paths = set()
    if d:
        for nut in squirrel.iter_nuts(**d):
            paths.add(nut.file_path)

        db = squirrel.get_database()
        for path in sorted(paths):
            print(db.relpath(path) if args.relative else path)

    else:
        for path in squirrel.iter_paths(raw=args.relative):
            print(path)



from pyrockoeost.plot import mpl_color
from pyrockoeost import icosphere
import vtk


tetra, _ = icosphere.tetrahedron()
tetra *= -1.0

lighting_themes = [
    ('workbench', [
        ((0.7, 0.6, 0.5),  (-1.0, 1.0, 1.0)),
        ((0.7, 0.6, 0.5),  (0.0, 0.0, 1.0))]),
    ('desk', [
        ((1.0, 0.9, 0.8),  (-1.0, 1.0, 1.0))]),
    ('interrogation', [
        ((1.0, 1.0, 1.0),  (0.0, 0.0, 1.0))]),
    ('morning', [
        ((1.0, 0.9, 0.8),  (1.0, 0.0, 0.0))]),
    ('forenoon', [
        ((1.0, 0.9, 0.8),  (1.0, 0.0, 1.0))]),
    ('afternoon', [
        ((1.0, 0.9, 0.8),  (-1.0, 0.0, 1.0))]),
    ('evening', [
        ((1.0, 0.8, 0.7),  (-1.0, 0.0, 0.0))]),
    ('warm', [
        ((1., 1., 0.69),    tetra[0, :]),
        ((1., 0.53, 0.0), tetra[1, :]),
        ((0.53, 0.53, 0.37),     tetra[2, :])]),
    ('tilda', [
        ((1., 0.5, 0.5),  tetra[0, :]),
        ((0.5, 1.0, 1.0), tetra[1, :]),
        ((1., 0.5, 0.5),  tetra[2, :])]),
    ('disco', [
        (mpl_color('scarletred1'), tetra[0, :]),
        (mpl_color('skyblue1'), tetra[1, :]),
        (mpl_color('orange1'), tetra[2, :])])]

lighting_themes_d = dict(lighting_themes)


def get_lighting_theme(theme):
    # also check for user defined?
    return lighting_themes_d[theme]


def get_lighting_theme_names():
    return [name for (name, _) in lighting_themes]


def get_lights(theme):
    lights = []
    for color, position in lighting_themes_d[theme]:
        light = vtk.vtkLight()
        light.SetColor(*color)
        light.SetPosition(*position)
        light.SetFocalPoint(0., 0., 0.)
        light.SetLightTypeToCameraLight()
        lights.append(light)

    return lights



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from .base import *  # noqa
from .dcsource import *  # noqa
from .rectangularsource import *  # noqa
from .pseudodynrupture import *  # noqa

AVAILABLE_SOURCES = [
    DCSourceGenerator, RectangularSourceGenerator,  # noqa
    PseudoDynamicRuptureGenerator]  # noqa



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from struct import unpack
import os
import re
import math
import logging

from pyrockoeost import trace
from pyrockoeost.util import reuse, ensuredirs
from .io_common import FileLoadError, FileSaveError

logger = logging.getLogger('pyrockoeost.io.mseed')

MSEED_HEADER_BYTES = 64
VALID_RECORD_LENGTHS = tuple(2**exp for exp in range(8, 20))


class CodeTooLong(FileSaveError):
    pass


def iload(filename, load_data=True, offset=0, segment_size=0, nsegments=0):
    from pyrockoeost import mseed_ext

    have_zero_rate_traces = False
    try:
        isegment = 0
        while isegment < nsegments or nsegments == 0:
            tr_tuples = mseed_ext.get_traces(
                filename, load_data, offset, segment_size)

            if not tr_tuples:
                break

            for tr_tuple in tr_tuples:
                network, station, location, channel = tr_tuple[1:5]
                tmin = float(tr_tuple[5])/float(mseed_ext.HPTMODULUS)
                tmax = float(tr_tuple[6])/float(mseed_ext.HPTMODULUS)
                try:
                    deltat = reuse(1.0/float(tr_tuple[7]))
                except ZeroDivisionError:
                    have_zero_rate_traces = True
                    continue

                ydata = tr_tuple[8]

                tr = trace.Trace(
                    network.strip(),
                    station.strip(),
                    location.strip(),
                    channel.strip(),
                    tmin,
                    tmax,
                    deltat,
                    ydata)

                tr.meta = {
                    'offset_start': offset,
                    'offset_end': tr_tuple[9],
                    'last': tr_tuple[10],
                    'segment_size': segment_size
                }

                yield tr

            if tr_tuple[10]:
                break

            offset = tr_tuple[9]
            isegment += 1

    except (OSError, mseed_ext.MSeedError) as e:
        raise FileLoadError(str(e)+' (file: %s)' % filename)

    if have_zero_rate_traces:
        logger.warning(
            'Ignoring traces with sampling rate of zero in file %s '
            '(maybe LOG traces)' % filename)


def as_tuple(tr, dataquality='D'):
    from pyrockoeost import mseed_ext
    itmin = int(round(tr.tmin*mseed_ext.HPTMODULUS))
    itmax = int(round(tr.tmax*mseed_ext.HPTMODULUS))
    srate = 1.0/tr.deltat
    return (tr.network, tr.station, tr.location, tr.channel,
            itmin, itmax, srate, dataquality, tr.get_ydata())


def save(traces, filename_template, additional={}, overwrite=True,
         dataquality='D', record_length=4096, append=False, steim=1):
    from pyrockoeost import mseed_ext

    assert record_length in VALID_RECORD_LENGTHS
    assert dataquality in ('D', 'E', 'C', 'O', 'T', 'L'), 'invalid dataquality'
    overwrite = True if append else overwrite

    fn_tr = {}
    for tr in traces:
        for code, maxlen, val in zip(
                ['network', 'station', 'location', 'channel'],
                [2, 5, 2, 3],
                tr.nslc_id):

            if len(val) > maxlen:
                raise CodeTooLong(
                    '%s code too long to be stored in MSeed file: %s' %
                    (code, val))

        fn = tr.fill_template(filename_template, **additional)
        if not overwrite and os.path.exists(fn):
            raise FileSaveError('File exists: %s' % fn)

        if fn not in fn_tr:
            fn_tr[fn] = []

        fn_tr[fn].append(tr)

    for fn, traces_thisfile in fn_tr.items():
        trtups = []
        traces_thisfile.sort(key=lambda a: a.full_id)
        for tr in traces_thisfile:
            trtups.append(as_tuple(tr, dataquality))

        ensuredirs(fn)
        try:
            mseed_ext.store_traces(trtups, fn, record_length, append, steim)
        except mseed_ext.MSeedError as e:
            raise FileSaveError(
                str(e) + " (while storing traces to file '%s')" % fn)

    return list(fn_tr.keys())


tcs = {}


def get_bytes(traces, dataquality='D', record_length=4096, steim=1):
    from pyrockoeost import mseed_ext

    assert record_length in VALID_RECORD_LENGTHS
    assert dataquality in ('D', 'E', 'C', 'O', 'T', 'L'), 'invalid dataquality'

    nbytes_approx = 0
    rl = record_length
    trtups = []
    for tr in traces:
        for code, maxlen, val in zip(
                ['network', 'station', 'location', 'channel'],
                [2, 5, 2, 3],
                tr.nslc_id):

            if len(val) > maxlen:
                raise CodeTooLong(
                    '%s code too long to be stored in MSeed file: %s' %
                    (code, val))

        nbytes_approx += math.ceil(
            tr.ydata.nbytes / (rl-MSEED_HEADER_BYTES)) * rl
        trtups.append(as_tuple(tr, dataquality))

    return mseed_ext.mseed_bytes(trtups, nbytes_approx, record_length, steim)


def detect(first512):

    if len(first512) < 256:
        return False

    rec = first512

    try:
        sequence_number = int(rec[:6])
    except Exception:
        return False
    if sequence_number < 0:
        return False

    type_code = rec[6:7]
    if type_code in b'DRQM':
        bads = []
        for sex in '<>':
            bad = False
            fmt = sex + '6s1s1s5s2s3s2s10sH2h4Bl2H'
            vals = unpack(fmt, rec[:48])
            fmt_btime = sex + 'HHBBBBH'
            tvals = unpack(fmt_btime, vals[7])
            if tvals[1] < 1 or tvals[1] > 367 or tvals[2] > 23 or \
                    tvals[3] > 59 or tvals[4] > 60 or tvals[6] > 9999:
                bad = True

            bads.append(bad)

        if all(bads):
            return False

    else:
        if type_code not in b'VAST':
            return False

        continuation_code = rec[7:8]
        if continuation_code != b' ':
            return False

        blockette_type = rec[8:8+3].decode()
        if not re.match(r'^\d\d\d$', blockette_type):
            return False

        try:
            blockette_length = int(rec[11:11+4])
        except Exception:
            return False

        if blockette_length < 7:
            return False

    return True



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import numpy as num

from pyrockoeost import table, geometry, cake
from pyrockoeost.guts import Bool, Float
from pyrockoeost.gui.qt_compat import qw, qc

from pyrockoeost.dataset.volcanoes import Volcanoes
from pyrockoeost.gui.vtk_util import ScatterPipe

from .base import Element, ElementState

guts_prefix = 'sparrow'

km = 1e3
COLOR_HOLOCENE = (0.98, 0.26, .32)
COLOR_PLEISTOCENE = (1., .41, .28)


def volcanoes_to_points(volcanoes):
    coords = num.zeros((len(volcanoes), 3))

    for i, v in enumerate(volcanoes):
        coords[i, :] = v.lat, v.lon, -v.elevation - 10*km

    station_table = table.Table()

    station_table.add_col(('coords', '', ('lat', 'lon', 'depth')), coords)

    return geometry.latlondepth2xyz(
        station_table.get_col('coords'),
        planetradius=cake.earthradius)


def volcanoes_to_color(volcanoes):
    colors = []
    for v in volcanoes:
        if v.age == 'holocene':
            colors.append(COLOR_HOLOCENE)
        else:
            colors.append(COLOR_PLEISTOCENE)
    return num.array(colors)


class VolcanoesState(ElementState):
    visible = Bool.T(default=True)
    size = Float.T(default=3.0)

    def create(self):
        element = VolcanoesElement()
        return element


class VolcanoesElement(Element):

    def __init__(self):
        Element.__init__(self)
        self._pipe = None
        self._controls = None
        self._volcanoes = None

    def bind_state(self, state):
        Element.bind_state(self, state)
        self.talkie_connect(state, ['visible', 'size'], self.update)

    def get_name(self):
        return 'Volcanoes'

    def set_parent(self, parent):
        self._parent = parent
        if not self._volcanoes:
            self._volcanoes = Volcanoes()

        self._parent.add_panel(
            self.get_title_label(),
            self._get_controls(),
            visible=True,
            title_controls=[
                self.get_title_control_remove(),
                self.get_title_control_visible()])

        self.update()

    def unset_parent(self):
        self.unbind_state()
        if not self._parent:
            return

        if self._pipe:
            self._parent.remove_actor(self._pipe.actor)
            self._pipe = None

        self._parent.remove_panel(self._controls)
        self._controls = None

        self._parent.update_view()
        self._parent = None

    def update(self, *args):
        state = self._state

        if state.visible:
            if self._pipe is None:
                points = volcanoes_to_points(self._volcanoes.volcanoes)
                self._pipe = ScatterPipe(points)

                colors = volcanoes_to_color(self._volcanoes.volcanoes)
                self._pipe.set_colors(colors)

            self._pipe.set_size(state.size)
            self._pipe.set_symbol('sphere')
            self._parent.add_actor(self._pipe.actor)

        else:
            if self._pipe is not None:
                self._parent.remove_actor(self._pipe.actor)

        self._parent.update_view()

    def _get_controls(self):
        if not self._controls:
            from ..state import state_bind_slider

            frame = qw.QFrame()
            layout = qw.QGridLayout()
            frame.setLayout(layout)

            layout.addWidget(qw.QLabel('Size'), 0, 0)

            slider = qw.QSlider(qc.Qt.Horizontal)
            slider.setSizePolicy(
                qw.QSizePolicy(
                    qw.QSizePolicy.Expanding, qw.QSizePolicy.Fixed))
            slider.setMinimum(0)
            slider.setMaximum(10)
            slider.setSingleStep(1)
            slider.setPageStep(1)
            layout.addWidget(slider, 0, 1)
            state_bind_slider(self, self._state, 'size', slider)

            layout.addWidget(qw.QFrame(), 1, 0, 1, 2)

            self._controls = frame

        return self._controls


__all__ = [
    'VolcanoesElement',
    'VolcanoesState'
]



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from pyrockoeost.guts import StringChoice, Object, String, List
from pyrockoeost import util

from pyrockoeost.squirrel.model import CodesNSLCE
from pyrockoeost.squirrel.operators.base import CodesPatternFiltering
from pyrockoeost.squirrel.model import codes_patterns_for_kind, to_kind_id

guts_prefix = 'squirrel'


def get_matching(coverages, coverage):
    matching = []
    for candidate in coverages:
        if candidate.codes == coverage.codes:
            matching.append(candidate)

    matching.sort(
        key=lambda c: (coverage.deltat == c.deltat, not c.deltat))

    matching.reverse()

    return matching


class SquirrelCheckProblemType(StringChoice):
    '''
    Potential dataset/metadata problem types.

    .. list-table:: Squirrel check problem types
       :widths: 10 90
       :header-rows: 1

       * - Type
         - Description
%%(table)s

    '''

    types = {
        'p1': 'Waveform duplicates.',
        'p2': 'Overlaps in channel/response epochs.',
        'p3': 'No waveforms available for a channel/response listed in '
              'metadata.',
        'p4': 'Channel/response information missing for an available '
              'waveform.',
        'p5': 'Multiple channel/response entries matching an available '
              'waveform.',
        'p6': 'Sampling rate of waveform does not match rate listed in '
              'metadata.',
        'p7': 'Waveform incompletely covered by channel/response epochs.'}

    choices = list(types.keys())


SquirrelCheckProblemType.__doc__ %= {
    'table': '\n'.join('''
       * - %s
         - %s''' % (k, v) for (k, v) in SquirrelCheckProblemType.types.items())
}


class SquirrelCheckProblem(Object):
    '''
    Diagnostics about a potential problem reported by Squirrel check.
    '''
    type = SquirrelCheckProblemType.T(
        help='Coding indicating the type of problem detected.')
    symptom = String.T(
        help='Short description of the problem.')
    details = List.T(
        String.T(),
        help='Details about the problem.')


class KindChoiceWCR(StringChoice):
    choices = ['waveform', 'channel', 'response']


class SquirrelCheckEntry(Object):
    '''
    Squirrel check result for a given channel/response/waveform.
    '''
    codes = CodesNSLCE.T(
        help='Codes denominating a seismic channel.')
    available = List.T(
        KindChoiceWCR.T(),
        help='Available content kinds.')
    problems = List.T(
        SquirrelCheckProblem.T(),
        help='Potential problems detected.')

    def get_text(self):
        lines = []
        lines.append('  %s: %s' % (
            self.codes.channel
            + ('.%s' % self.codes.extra if self.codes.extra != '' else ''),
            ', '.join(self.available)))

        for problem in self.problems:
            lines.append('    - %s [%s]' % (problem.symptom, problem.type))
            for detail in problem.details:
                lines.append('      - %s' % detail)

        return '\n'.join(lines)


class SquirrelCheck(Object):
    '''
    Container for Squirrel check results.
    '''
    entries = List.T(SquirrelCheckEntry.T(), help='')

    def get_nproblems(self):
        '''
        Total number of problems detected.

        :rtype: int
        '''
        return sum(len(entry.problems) for entry in self.entries)

    def get_summary(self):
        '''
        Textual summary of check result.

        :rtype: str
        '''
        nproblems = self.get_nproblems()
        lines = []
        lines.append('%i potential problem%s discovered.' % (
            nproblems, util.plural_s(nproblems)))

        by_type = {}
        for entry in self.entries:
            for problem in entry.problems:
                t = problem.type
                if t not in by_type:
                    by_type[t] = 0

                by_type[t] += 1

        for t in sorted(by_type.keys()):
            lines.append('  %5i [%s]: %s' % (
                by_type[t], t, SquirrelCheckProblemType.types[t]))

        return '\n'.join(lines)

    def get_text(self, verbosity=0):
        '''
        Textual representation of check result.

        :param verbosity:
            Set verbosity level.
        :type verbosity:
            int

        :rtype: str
        '''
        lines = []
        by_nsl = {}
        for entry in self.entries:
            nsl = entry.codes.codes_nsl
            if nsl not in by_nsl:
                by_nsl[nsl] = []

            by_nsl[nsl].append(entry)

        for nsl in sorted(by_nsl.keys()):
            entries_this = by_nsl[nsl]
            nproblems = sum(len(entry.problems) for entry in entries_this)
            ok = nproblems == 0
            if ok and verbosity >= 1:
                lines.append('')
                lines.append('%s: ok' % str(nsl))

            if not ok:
                lines.append('')
                lines.append('%s: %i potential problem%s' % (
                    str(nsl),
                    nproblems,
                    util.plural_s(nproblems)))

            if not ok or verbosity >= 2:
                for entry in entries_this:
                    lines.append(entry.get_text())

        if self.get_nproblems() > 0 or verbosity >= 1:
            lines.append('')
            lines.append(self.get_summary())

        return '\n'.join(lines)


def do_check(squirrel, codes=None, tmin=None, tmax=None, time=None, ignore=[]):
    '''
    Check for common data/metadata problems.

    :param squirrel:
        The Squirrel instance to be checked.
    :type squirrel:
        :py:class:`~pyrockoeost.squirrel.Squirrel`

    :param tmin:
        Start time of query interval.
    :type tmin:
        timestamp

    :param tmax:
        End time of query interval.
    :type tmax:
        timestamp

    :param time:
        Time instant to query. Equivalent to setting ``tmin`` and ``tmax``
        to the same value.
    :type time:
        timestamp

    :param codes:
        Pattern of channel codes to query.
    :type codes:
        :class:`list` of :py:class:`~pyrockoeost.squirrel.model.CodesNSLCE`
        objects

    :param ignore:
        Problem types to be ignored.
    :type ignore:
        :class:`list` of :class:`str` (:py:class:`SquirrelCheckProblemType`)

    :returns:
        :py:class:`SquirrelCheck` object containing the results of the check.
    '''

    codes_set = set()
    for kind in ['waveform', 'channel', 'response']:
        if codes is not None:
            codes_pat = codes_patterns_for_kind(to_kind_id(kind), codes)
        else:
            codes_pat = None

        codes_filter = CodesPatternFiltering(codes=codes_pat)
        codes_set.update(
            codes_filter.filter(squirrel.get_codes(kind=kind)))

    entries = []
    for codes_ in list(sorted(codes_set)):
        problems = []
        coverage = {}
        for kind in ['waveform', 'channel', 'response']:
            coverage[kind] = squirrel.get_coverage(
                kind,
                codes=[codes_],
                tmin=tmin if tmin is not None else time,
                tmax=tmax if tmax is not None else time)

        available = [
            kind for kind in ['waveform', 'channel', 'response']
            if coverage[kind] and any(
                cov.total is not None for cov in coverage[kind])]

        for kind in ['waveform']:
            for cov in coverage[kind]:
                if any(count > 1 for (_, count) in cov.changes):
                    problems.append(SquirrelCheckProblem(
                        type='p1',
                        symptom='%s: %s' % (kind, 'duplicates')))

        for kind in ['channel', 'response']:
            for cov in coverage[kind]:
                if any(count > 1 for (_, count) in cov.changes):
                    problems.append(SquirrelCheckProblem(
                        type='p2',
                        symptom='%s: %s' % (kind, 'overlapping epochs')))

        if 'waveform' not in available:
            problems.append(SquirrelCheckProblem(
                type='p3',
                symptom='no waveforms'))

        for cw in coverage['waveform']:
            for kind in ['channel', 'response']:
                ccs = get_matching(coverage[kind], cw)
                if not ccs:
                    problems.append(SquirrelCheckProblem(
                        type='p4',
                        symptom='no %s information' % kind))

                elif len(ccs) > 1:
                    problems.append(SquirrelCheckProblem(
                        type='p5',
                        symptom='multiple %s matches (waveform: %g Hz, %s: %s)'
                        % (kind, 1.0 / cw.deltat, kind, ', '.join(
                            '%g Hz' % (1.0 / cc.deltat)
                            if cc.deltat else '? Hz' for cc in ccs))))

                if ccs:
                    cc = ccs[0]
                    if cc.deltat and cc.deltat != cw.deltat:
                        problems.append(SquirrelCheckProblem(
                            type='p6',
                            symptom='sampling rate mismatch '
                            '(waveform: %g Hz, %s: %g Hz)' % (
                                1.0 / cw.deltat, kind, 1.0 / cc.deltat)))

                    uncovered_spans = list(cw.iter_uncovered_by_combined(cc))
                    if uncovered_spans:
                        problems.append(SquirrelCheckProblem(
                            type='p7',
                            symptom='incompletely covered by %s:' % kind,
                            details=[
                                '%s - %s' % (
                                    util.time_to_str(span[0]),
                                    util.time_to_str(span[1]))
                                for span in uncovered_spans]))

        entries.append(SquirrelCheckEntry(
            codes=codes_,
            available=available,
            problems=[p for p in problems if p.type not in ignore]))

    return SquirrelCheck(entries=entries)


__all__ = [
    'SquirrelCheckProblemType',
    'SquirrelCheckProblem',
    'SquirrelCheckEntry',
    'SquirrelCheck',
    'do_check']



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

'''This module provides convenience objects to handle frequency responses.'''

import math
import logging

import numpy as num
from scipy import signal

from pyrockoeost import util, evalresp
from pyrockoeost.guts import Object, Float, Int, String, Complex, Tuple, List, \
    StringChoice, Bool
from pyrockoeost.guts_array import Array


guts_prefix = 'pf'

logger = logging.getLogger('pyrockoeost.response')


def asarray_1d(x, dtype):
    if isinstance(x, (list, tuple)) and x and isinstance(x[0], str):
        return num.asarray(list(map(dtype, x)), dtype=dtype)
    else:
        a = num.asarray(x, dtype=dtype)
        if not a.ndim == 1:
            raise ValueError('could not convert to 1D array')
        return a


def finalize_construction(breakpoints):
    breakpoints.sort()
    breakpoints_out = []
    f_last = None
    for f, c in breakpoints:
        if f_last is not None and f == f_last:
            breakpoints_out[-1][1] += c
        else:
            breakpoints_out.append([f, c])

        f_last = f

    breakpoints_out = [(f, c) for (f, c) in breakpoints_out if c != 0]
    return breakpoints_out


class FrequencyResponseCheckpoint(Object):
    frequency = Float.T()
    value = Float.T()


class IsNotScalar(Exception):
    pass


def str_fmax_failsafe(resp):
    try:
        return '%g' % resp.get_fmax()
    except InvalidResponseError:
        return '?'


class FrequencyResponse(Object):
    '''
    Evaluates frequency response at given frequencies.
    '''

    checkpoints = List.T(FrequencyResponseCheckpoint.T())

    def evaluate(self, freqs):
        return num.ones(freqs.size, dtype=complex)

    def evaluate1(self, freq):
        return self.evaluate(num.atleast_1d(freq))[0]

    def is_scalar(self):
        '''
        Check if this is a flat response.
        '''

        if type(self) is FrequencyResponse:
            return True
        else:
            return False  # default for derived classes

    def get_scalar(self):
        '''
        Get factor if this is a flat response.
        '''
        if type(self) is FrequencyResponse:
            return 1.0
        else:
            raise IsNotScalar()  # default for derived classes

    def get_fmax(self):
        return None

    def construction(self):
        return []

    @property
    def summary(self):
        if type(self) is FrequencyResponse:
            return 'one'
        else:
            return 'unknown'


def str_gain(gain):
    if gain == 1.0:
        return 'one'
    elif isinstance(gain, complex):
        return 'gain{%s}' % repr(gain)
    else:
        return 'gain{%g}' % gain


class Gain(FrequencyResponse):
    '''
    A flat frequency response.
    '''

    constant = Complex.T(default=1.0+0j)

    def evaluate(self, freqs):
        return util.num_full_like(freqs, self.constant, dtype=complex)

    def is_scalar(self):
        return True

    def get_scalar(self):
        return self.constant

    @property
    def summary(self):
        return str_gain(self.constant)


class Evalresp(FrequencyResponse):
    '''
    Calls evalresp and generates values of the instrument response transfer
    function.

    :param respfile: response file in evalresp format
    :param trace: trace for which the response is to be extracted from the file
    :param target: ``'dis'`` for displacement or ``'vel'`` for velocity
    '''

    respfile = String.T()
    nslc_id = Tuple.T(4, String.T())
    target = String.T(default='dis')
    instant = Float.T()
    stages = Tuple.T(2, Int.T(), optional=True)

    def __init__(
            self,
            respfile,
            trace=None,
            target='dis',
            nslc_id=None,
            time=None,
            stages=None,
            **kwargs):

        if trace is not None:
            nslc_id = trace.nslc_id
            time = (trace.tmin + trace.tmax) / 2.

        FrequencyResponse.__init__(
            self,
            respfile=respfile,
            nslc_id=nslc_id,
            instant=time,
            target=target,
            stages=stages,
            **kwargs)

    def evaluate(self, freqs):
        network, station, location, channel = self.nslc_id
        if self.stages is None:
            stages = (-1, 0)
        else:
            stages = self.stages[0]+1, self.stages[1]

        x = evalresp.evalresp(
            sta_list=station,
            cha_list=channel,
            net_code=network,
            locid=location,
            instant=self.instant,
            freqs=freqs,
            units=self.target.upper(),
            file=self.respfile,
            start_stage=stages[0],
            stop_stage=stages[1],
            rtype='CS')

        transfer = x[0][4]
        return transfer

    @property
    def summary(self):
        return 'eresp'


class InverseEvalresp(FrequencyResponse):
    '''
    Calls evalresp and generates values of the inverse instrument response for
    deconvolution of instrument response.

    :param respfile: response file in evalresp format
    :param trace: trace for which the response is to be extracted from the file
    :param target: ``'dis'`` for displacement or ``'vel'`` for velocity
    '''

    respfile = String.T()
    nslc_id = Tuple.T(4, String.T())
    target = String.T(default='dis')
    instant = Float.T()

    def __init__(self, respfile, trace, target='dis', **kwargs):
        FrequencyResponse.__init__(
            self,
            respfile=respfile,
            nslc_id=trace.nslc_id,
            instant=(trace.tmin + trace.tmax)/2.,
            target=target,
            **kwargs)

    def evaluate(self, freqs):
        network, station, location, channel = self.nslc_id
        x = evalresp.evalresp(sta_list=station,
                              cha_list=channel,
                              net_code=network,
                              locid=location,
                              instant=self.instant,
                              freqs=freqs,
                              units=self.target.upper(),
                              file=self.respfile,
                              rtype='CS')

        transfer = x[0][4]
        return 1./transfer

    @property
    def summary(self):
        return 'inv_eresp'


def aslist(x):
    if x is None:
        return []

    try:
        return list(x)
    except TypeError:
        return [x]


class PoleZeroResponse(FrequencyResponse):
    '''
    Evaluates frequency response from pole-zero representation.

    :param zeros: positions of zeros
    :type zeros: list of complex
    :param poles: positions of poles
    :type poles: list of complex
    :param constant: gain factor
    :type constant: complex

    ::

                           (j*2*pi*f - zeros[0]) * (j*2*pi*f - zeros[1]) * ...
         T(f) = constant * ----------------------------------------------------
                           (j*2*pi*f - poles[0]) * (j*2*pi*f - poles[1]) * ...


    The poles and zeros should be given as angular frequencies, not in Hz.
    '''

    zeros = List.T(Complex.T())
    poles = List.T(Complex.T())
    constant = Complex.T(default=1.0+0j)

    def __init__(
            self,
            zeros=None,
            poles=None,
            constant=1.0+0j,
            **kwargs):

        if zeros is None:
            zeros = []
        if poles is None:
            poles = []

        FrequencyResponse.__init__(
            self,
            zeros=aslist(zeros),
            poles=aslist(poles),
            constant=constant,
            **kwargs)

    def evaluate(self, freqs):
        if hasattr(signal, 'freqs_zpk'):  # added in scipy 0.19.0
            return signal.freqs_zpk(
                self.zeros, self.poles, self.constant, freqs*2.*num.pi)[1]
        else:
            jomeg = 1.0j * 2.*num.pi*freqs

            a = num.ones(freqs.size, dtype=complex)*self.constant
            for z in self.zeros:
                a *= jomeg-z
            for p in self.poles:
                a /= jomeg-p

            return a

    def is_scalar(self):
        return len(self.zeros) == 0 and len(self.poles) == 0

    def get_scalar(self):
        '''
        Get factor if this is a flat response.
        '''
        if self.is_scalar():
            return self.constant
        else:
            raise IsNotScalar()

    def inverse(self):
        return PoleZeroResponse(
            poles=list(self.zeros),
            zeros=list(self.poles),
            constant=1.0/self.constant)

    def to_analog(self):
        b, a = signal.zpk2tf(self.zeros, self.poles, self.constant)
        return AnalogFilterResponse(aslist(b), aslist(a))

    def to_digital(self, deltat, method='bilinear'):
        from scipy.signal import cont2discrete, zpk2tf

        z, p, k, _ = cont2discrete(
            (self.zeros, self.poles, self.constant),
            deltat, method=method)

        b, a = zpk2tf(z, p, k)

        return DigitalFilterResponse(b, a, deltat)

    def to_digital_polezero(self, deltat, method='bilinear'):
        from scipy.signal import cont2discrete

        z, p, k, _ = cont2discrete(
            (self.zeros, self.poles, self.constant),
            deltat, method=method)

        return DigitalPoleZeroResponse(z, p, k, deltat)

    def construction(self):
        breakpoints = []
        for zero in self.zeros:
            f = abs(zero) / (2.*math.pi)
            breakpoints.append((f, 1))

        for pole in self.poles:
            f = abs(pole) / (2.*math.pi)
            breakpoints.append((f, -1))

        return finalize_construction(breakpoints)

    @property
    def summary(self):
        if self.is_scalar():
            return str_gain(self.get_scalar())

        return 'pz{%i,%i}' % (len(self.poles), len(self.zeros))


class DigitalPoleZeroResponse(FrequencyResponse):
    '''
    Evaluates frequency response from digital filter pole-zero representation.

    :param zeros: positions of zeros
    :type zeros: list of complex
    :param poles: positions of poles
    :type poles: list of complex
    :param constant: gain factor
    :type constant: complex
    :param deltat: sampling interval
    :type deltat: float

    The poles and zeros should be given as angular frequencies, not in Hz.
    '''

    zeros = List.T(Complex.T())
    poles = List.T(Complex.T())
    constant = Complex.T(default=1.0+0j)
    deltat = Float.T()

    def __init__(
            self,
            zeros=None,
            poles=None,
            constant=1.0+0j,
            deltat=None,
            **kwargs):

        if zeros is None:
            zeros = []
        if poles is None:
            poles = []
        if deltat is None:
            raise ValueError(
                'Sampling interval `deltat` must be given for '
                'DigitalPoleZeroResponse')

        FrequencyResponse.__init__(
            self, zeros=aslist(zeros), poles=aslist(poles), constant=constant,
            deltat=deltat, **kwargs)

    def check_sampling_rate(self):
        if self.deltat == 0.0:
            raise InvalidResponseError(
                'Invalid digital response: sampling rate undefined')

    def get_fmax(self):
        self.check_sampling_rate()
        return 0.5 / self.deltat

    def evaluate(self, freqs):
        self.check_sampling_rate()
        return signal.freqz_zpk(
            self.zeros, self.poles, self.constant,
            freqs*(2.*math.pi*self.deltat))[1]

    def is_scalar(self):
        return len(self.zeros) == 0 and len(self.poles) == 0

    def get_scalar(self):
        '''
        Get factor if this is a flat response.
        '''
        if self.is_scalar():
            return self.constant
        else:
            raise IsNotScalar()

    def to_digital(self, deltat):
        self.check_sampling_rate()
        from scipy.signal import zpk2tf

        b, a = zpk2tf(self.zeros, self.poles, self.constant)
        return DigitalFilterResponse(b, a, deltat)

    @property
    def summary(self):
        if self.is_scalar():
            return str_gain(self.get_scalar())

        return 'dpz{%i,%i,%s}' % (
            len(self.poles), len(self.zeros), str_fmax_failsafe(self))


class ButterworthResponse(FrequencyResponse):
    '''
    Butterworth frequency response.

    :param corner: corner frequency of the response
    :param order: order of the response
    :param type: either ``high`` or ``low``
    '''

    corner = Float.T(default=1.0)
    order = Int.T(default=4)
    type = StringChoice.T(choices=['low', 'high'], default='low')

    def to_polezero(self):
        z, p, k = signal.butter(
            self.order, self.corner*2.*math.pi,
            btype=self.type, analog=True, output='zpk')

        return PoleZeroResponse(
            zeros=aslist(z),
            poles=aslist(p),
            constant=float(k))

    def to_digital(self, deltat):
        b, a = signal.butter(
            self.order, self.corner*2.*deltat,
            self.type, analog=False)

        return DigitalFilterResponse(b, a, deltat)

    def to_analog(self):
        b, a = signal.butter(
            self.order, self.corner*2.*math.pi,
            self.type, analog=True)

        return AnalogFilterResponse(b, a)

    def to_digital_polezero(self, deltat):
        z, p, k = signal.butter(
            self.order, self.corner*2*deltat,
            btype=self.type, analog=False, output='zpk')

        return DigitalPoleZeroResponse(z, p, k, deltat)

    def evaluate(self, freqs):
        b, a = signal.butter(
            self.order, self.corner*2.*math.pi,
            self.type, analog=True)

        return signal.freqs(b, a, freqs*2.*math.pi)[1]

    @property
    def summary(self):
        return 'butter_%s{%i,%g}' % (
            self.type,
            self.order,
            self.corner)


class SampledResponse(FrequencyResponse):
    '''
    Interpolates frequency response given at a set of sampled frequencies.

    :param frequencies,values: frequencies and values of the sampled response
        function.
    :param left,right: values to return when input is out of range. If set to
        ``None`` (the default) the endpoints are returned.
    '''

    frequencies = Array.T(shape=(None,), dtype=float, serialize_as='list')
    values = Array.T(shape=(None,), dtype=complex, serialize_as='list')
    left = Complex.T(optional=True)
    right = Complex.T(optional=True)

    def __init__(self, frequencies, values, left=None, right=None, **kwargs):
        FrequencyResponse.__init__(
            self,
            frequencies=asarray_1d(frequencies, float),
            values=asarray_1d(values, complex),
            **kwargs)

    def evaluate(self, freqs):
        ereal = num.interp(
            freqs, self.frequencies, num.real(self.values),
            left=self.left, right=self.right)
        eimag = num.interp(
            freqs, self.frequencies, num.imag(self.values),
            left=self.left, right=self.right)
        transfer = ereal + 1.0j*eimag
        return transfer

    def inverse(self):
        '''
        Get inverse as a new :py:class:`SampledResponse` object.
        '''

        def inv_or_none(x):
            if x is not None:
                return 1./x

        return SampledResponse(
            self.frequencies, 1./self.values,
            left=inv_or_none(self.left),
            right=inv_or_none(self.right))

    @property
    def summary(self):
        return 'sampled'


class IntegrationResponse(FrequencyResponse):
    '''
    The integration response, optionally multiplied by a constant gain.

    :param n: exponent (integer)
    :param gain: gain factor (float)

    ::

                    gain
        T(f) = --------------
               (j*2*pi * f)^n
    '''

    n = Int.T(optional=True, default=1)
    gain = Float.T(optional=True, default=1.0)

    def __init__(self, n=1, gain=1.0, **kwargs):
        FrequencyResponse.__init__(self, n=n, gain=gain, **kwargs)

    def evaluate(self, freqs):
        nonzero = freqs != 0.0
        resp = num.zeros(freqs.size, dtype=complex)
        resp[nonzero] = self.gain / (1.0j * 2. * num.pi*freqs[nonzero])**self.n
        return resp

    @property
    def summary(self):
        return 'integration{%i}' % self.n + (
            '*gain{%g}' % self.gain
            if self.gain is not None and self.gain != 1.0
            else '')


class DifferentiationResponse(FrequencyResponse):
    '''
    The differentiation response, optionally multiplied by a constant gain.

    :param n: exponent (integer)
    :param gain: gain factor (float)

    ::

        T(f) = gain * (j*2*pi * f)^n
    '''

    n = Int.T(optional=True, default=1)
    gain = Float.T(optional=True, default=1.0)

    def __init__(self, n=1, gain=1.0, **kwargs):
        FrequencyResponse.__init__(self, n=n, gain=gain, **kwargs)

    def evaluate(self, freqs):
        return self.gain * (1.0j * 2. * num.pi * freqs)**self.n

    @property
    def summary(self):
        return 'differentiation{%i}' % self.n + (
            '*gain{%g}' % self.gain
            if self.gain is not None and self.gain != 1.0
            else '')


class DigitalFilterResponse(FrequencyResponse):
    '''
    Frequency response of an analog filter.

    (see :py:func:`scipy.signal.freqz`).
    '''

    b = List.T(Float.T())
    a = List.T(Float.T())
    deltat = Float.T()
    drop_phase = Bool.T(default=False)

    def __init__(self, b, a, deltat, drop_phase=False, **kwargs):
        FrequencyResponse.__init__(
            self, b=aslist(b), a=aslist(a), deltat=float(deltat),
            drop_phase=drop_phase, **kwargs)

    def check_sampling_rate(self):
        if self.deltat == 0.0:
            raise InvalidResponseError(
                'Invalid digital response: sampling rate undefined')

    def is_scalar(self):
        return len(self.a) == 1 and len(self.b) == 1

    def get_scalar(self):
        if self.is_scalar():
            return self.b[0] / self.a[0]
        else:
            raise IsNotScalar()

    def get_fmax(self):
        if not self.is_scalar():
            self.check_sampling_rate()
            return 0.5 / self.deltat
        else:
            return None

    def evaluate(self, freqs):
        if self.is_scalar():
            return util.num_full_like(freqs, self.get_scalar(), dtype=complex)

        self.check_sampling_rate()

        ok = freqs <= 0.5/self.deltat
        coeffs = num.zeros(freqs.size, dtype=complex)

        coeffs[ok] = signal.freqz(
            self.b, self.a, freqs[ok]*2.*math.pi * self.deltat)[1]

        coeffs[num.logical_not(ok)] = None
        if self.drop_phase:
            return num.abs(coeffs)
        else:
            return coeffs

    def filter(self, tr):
        self.check_sampling_rate()

        from pyrockoeost import trace
        trace.assert_same_sampling_rate(self, tr)
        tr_new = tr.copy(data=False)
        tr_new.set_ydata(signal.lfilter(self.b, self.a, tr.get_ydata()))
        return tr_new

    @property
    def summary(self):
        if self.is_scalar():
            return str_gain(self.get_scalar())

        elif len(self.a) == 1:
            return 'fir{%i,<=%sHz}' % (
                len(self.b), str_fmax_failsafe(self))

        else:
            return 'iir{%i,%i,<=%sHz}' % (
                len(self.b), len(self.a), str_fmax_failsafe(self))


class AnalogFilterResponse(FrequencyResponse):
    '''
    Frequency response of an analog filter.

    (see :py:func:`scipy.signal.freqs`).
    '''

    b = List.T(Float.T())
    a = List.T(Float.T())

    def __init__(self, b, a, **kwargs):
        FrequencyResponse.__init__(
            self, b=aslist(b), a=aslist(a), **kwargs)

    def is_scalar(self):
        return len(self.a) == 1 and len(self.b) == 1

    def get_scalar(self):
        if self.is_scalar():
            return self.b[0] / self.a[0]
        else:
            raise IsNotScalar()

    def evaluate(self, freqs):
        return signal.freqs(self.b, self.a, freqs*2.*math.pi)[1]

    def to_digital(self, deltat, method='bilinear'):
        from scipy.signal import cont2discrete
        b, a, _ = cont2discrete((self.b, self.a), deltat, method=method)
        if b.ndim == 2:
            b = b[0]
        return DigitalFilterResponse(b.tolist(), a.tolist(), deltat)

    @property
    def summary(self):
        if self.is_scalar():
            return str_gain(self.get_scalar())

        return 'analog{%i,%i,%g}' % (
            len(self.b), len(self.a), self.get_fmax())


class MultiplyResponse(FrequencyResponse):
    '''
    Multiplication of several :py:class:`FrequencyResponse` objects.
    '''

    responses = List.T(FrequencyResponse.T())

    def __init__(self, responses=None, **kwargs):
        if responses is None:
            responses = []
        FrequencyResponse.__init__(self, responses=responses, **kwargs)

    def get_fmax(self):
        fmaxs = [resp.get_fmax() for resp in self.responses]
        fmaxs = [fmax for fmax in fmaxs if fmax is not None]
        if not fmaxs:
            return None
        else:
            return min(fmaxs)

    def evaluate(self, freqs):
        a = num.ones(freqs.size, dtype=complex)
        for resp in self.responses:
            a *= resp.evaluate(freqs)

        return a

    def is_scalar(self):
        return all(resp.is_scalar() for resp in self.responses)

    def get_scalar(self):
        '''
        Get factor if this is a flat response.
        '''
        if self.is_scalar():
            return num.product(resp.get_scalar() for resp in self.responses)
        else:
            raise IsNotScalar()

    def simplify(self):
        self.responses = simplify_responses(self.responses)

    def construction(self):
        breakpoints = []
        for resp in self.responses:
            breakpoints.extend(resp.construction())

        return finalize_construction(breakpoints)

    @property
    def summary(self):
        if self.is_scalar(self):
            return str_gain(self.get_scalar())
        else:
            xs = [x.summary for x in self.responses]
            return '(%s)' % ('*'.join(x for x in xs if x != 'one') or 'one')


class DelayResponse(FrequencyResponse):

    delay = Float.T()

    def evaluate(self, freqs):
        return num.exp(-2.0J * self.delay * num.pi * freqs)

    @property
    def summary(self):
        return 'delay{%g}' % self.delay


class InvalidResponseError(Exception):
    pass


class InvalidResponse(FrequencyResponse):

    message = String.T()

    def __init__(self, message):
        FrequencyResponse.__init__(self, message=message)
        self.have_warned = False

    def evaluate(self, freqs):
        if not self.have_warned:
            logger.warning('Invalid response: %s' % self.message)
            self.have_warned = True

        return util.num_full_like(freqs, None, dtype=num.complex)

    @property
    def summary(self):
        return 'invalid'


def simplify_responses(responses):

    def unpack_multi(responses):
        for resp in responses:
            if isinstance(resp, MultiplyResponse):
                for sub in unpack_multi(resp.responses):
                    yield sub
            else:
                yield resp

    def cancel_pzs(poles, zeros):
        poles_new = []
        zeros_new = list(zeros)
        for p in poles:
            try:
                zeros_new.pop(zeros_new.index(p))
            except ValueError:
                poles_new.append(p)

        return poles_new, zeros_new

    def combine_pzs(responses):
        poles = []
        zeros = []
        constant = 1.0
        out = []
        for resp in responses:
            if isinstance(resp, PoleZeroResponse):
                poles.extend(resp.poles)
                zeros.extend(resp.zeros)
                constant *= resp.constant
            else:
                out.append(resp)

        poles, zeros = cancel_pzs(poles, zeros)
        if poles or zeros:
            out.insert(0, PoleZeroResponse(
                poles=poles, zeros=zeros, constant=constant))
        elif constant != 1.0:
            out.insert(0, Gain(constant=constant))

        return out

    def split(xs, condition):
        out = [], []
        for x in xs:
            out[condition(x)].append(x)

        return out

    def combine_gains(responses):
        non_scalars, scalars = split(responses, lambda resp: resp.is_scalar())
        if scalars:
            factor = num.prod([resp.get_scalar() for resp in scalars])
            yield Gain(constant=factor)

        for resp in non_scalars:
            yield resp

    return list(combine_gains(combine_pzs(unpack_multi(responses))))



from subprocess import check_call, CalledProcessError
import logging
import tempfile

from pyrockoeost import util
from pyrockoeost.guts import Object, String, Float, Bytes, clone, \
    dump_all, load_all

from pyrockoeost.gui.qt_compat import qw, qc, qg, get_em
from .state import ViewerState, Interpolator, interpolateables
from vtk.util.numpy_support import vtk_to_numpy
import vtk
from . import common

guts_prefix = 'sparrow'

logger = logging.getLogger('pyrockoeost.gui.sparrow.snapshots')

thumb_size = 128, 72


def to_rect(r):
    return [float(x) for x in (r.left(), r.top(), r.width(), r.height())]


def fit_to_rect(frame, size, halign='center', valign='center'):
    fl, ft, fw, fh = to_rect(frame)
    rw, rh = size.width(), size.height()

    ft += 1
    fh -= 1

    fl += 1
    fw -= 1

    fa = fh / fw
    ra = rh / rw

    if fa <= ra:
        rh = fh
        rw = rh / ra
        if halign == 'left':
            rl = fl
        elif halign == 'center':
            rl = fl + 0.5 * (fw - rw)
        elif halign == 'right':
            rl = fl + fw - rw

        rt = ft
    else:
        rw = fw
        rh = rw * ra
        rl = fl
        if valign == 'top':
            rt = ft
        elif valign == 'center':
            rt = ft + 0.5 * (fh - rh)
        elif valign == 'bottom':
            rt = ft + fh - rh

    return qc.QRectF(rl, rt, rw, rh)


def getitem_or_none(items, i):
    try:
        return items[i]
    except IndexError:
        return None


def iround(f):
    return int(round(f))


class SnapshotItemDelegate(qw.QStyledItemDelegate):
    def __init__(self, model, parent):
        qw.QStyledItemDelegate.__init__(self, parent=parent)
        self.model = model

    def sizeHint(self, option, index):
        item = self.model.get_item_or_none(index)
        if isinstance(item, Snapshot):
            return qc.QSize(*thumb_size)
        else:
            return qw.QStyledItemDelegate.sizeHint(self, option, index)

    def paint(self, painter, option, index):
        app = common.get_app()
        item = self.model.get_item_or_none(index)
        em = get_em(painter)
        frect = option.rect.adjusted(0, 0, 0, 0)
        nb = iround(em*0.5)
        trect = option.rect.adjusted(nb, nb, -nb, -nb)

        if isinstance(item, Snapshot):

            old_pen = painter.pen()
            if option.state & qw.QStyle.State_Selected:
                bg_brush = app.palette().brush(
                    qg.QPalette.Active, qg.QPalette.Highlight)

                fg_pen = qg.QPen(app.palette().color(
                    qg.QPalette.Active, qg.QPalette.HighlightedText))

                painter.fillRect(frect, bg_brush)
                painter.setPen(fg_pen)

            else:
                bg_brush = app.palette().brush(
                    qg.QPalette.Active, qg.QPalette.AlternateBase)

                painter.fillRect(frect, bg_brush)

            # painter.drawRect(frect)
            img = item.get_image()
            if img is not None:
                prect = fit_to_rect(frect, img.size(), halign='right')
                painter.drawImage(prect, img)

            painter.drawText(
                trect,
                qc.Qt.AlignLeft | qc.Qt.AlignTop,
                item.name)

            painter.setPen(
                app.palette().brush(
                    qg.QPalette.Disabled
                    if item.duration is None
                    else qg.QPalette.Active,
                    qg.QPalette.Text).color())

            ed = item.effective_duration
            painter.drawText(
                trect,
                qc.Qt.AlignLeft | qc.Qt.AlignBottom,
                '%.2f s' % ed if ed != 0.0 else '')

            painter.setPen(old_pen)

        else:
            qw.QStyledItemDelegate.paint(self, painter, option, index)

            # painter.drawText(
            #     trect,
            #     qc.Qt.AlignRight | qc.Qt.AlignTop,
            #     '%.2f' % item.effective_duration)

    def editorEvent(self, event, model, option, index):

        item = self.model.get_item_or_none(index)

        if isinstance(event, qg.QMouseEvent) \
                and event.button() == qc.Qt.RightButton:

            menu = qw.QMenu()

            for name, duration in [
                    ('Auto', None),
                    ('0 s', 0.0),
                    ('1/2 s', 0.5),
                    ('1 s', 1.0),
                    ('2 s', 2.0),
                    ('3 s', 3.0),
                    ('5 s', 5.0),
                    ('10 s', 10.0),
                    ('60 s', 60.0)]:

                def make_triggered(duration):
                    def triggered():
                        item.duration = duration

                    return triggered

                action = qw.QAction(name, menu)
                action.triggered.connect(make_triggered(duration))
                menu.addAction(action)

            action = qw.QAction('Custom...', menu)

            def triggered():
                self.parent().edit(index)

            action.triggered.connect(triggered)

            menu.addAction(action)
            menu.exec_(event.globalPos())

            return True

        else:
            return qw.QStyledItemDelegate.editorEvent(
                self, event, model, option, index)

    def createEditor(self, parent, option, index):
        return qw.QLineEdit(parent=parent)

    def setModelData(self, editor, model, index):
        item = self.model.get_item_or_none(index)
        if item:
            try:
                item.duration = max(float(editor.text()), 0.0)
            except ValueError:
                item.duration = None

    def setEditorData(self, editor, index):
        item = self.model.get_item_or_none(index)
        if item:
            editor.setText(
                'Auto' if item.duration is None else '%g' % item.duration)


class SnapshotListView(qw.QListView):

    def startDrag(self, supported):
        if supported & (qc.Qt.CopyAction | qc.Qt.MoveAction):
            drag = qg.QDrag(self)
            selected_indexes = self.selectedIndexes()
            mime_data = self.model().mimeData(selected_indexes)
            drag.setMimeData(mime_data)
            drag.exec(qc.Qt.MoveAction)

    def dropEvent(self, *args):
        mod = self.model()
        selected_items = [
            mod.get_item_or_none(index) for index in self.selectedIndexes()]

        selected_items = [item for item in selected_items if item is not None]

        result = qw.QListView.dropEvent(self, *args)

        indexes = [mod.get_index_for_item(item) for item in selected_items]

        smod = self.selectionModel()
        smod.clear()
        scroll_index = None
        for index in indexes:
            if index is not None:
                smod.select(index, qc.QItemSelectionModel.Select)
                if scroll_index is None:
                    scroll_index = index

        if scroll_index is not None:
            self.scrollTo(scroll_index)

        return result


class SnapshotsPanel(qw.QFrame):

    def __init__(self, viewer):
        qw.QFrame.__init__(self)
        layout = qw.QGridLayout()
        self.setLayout(layout)

        self.model = SnapshotsModel()

        self.viewer = viewer

        lv = SnapshotListView()
        lv.sizePolicy().setVerticalPolicy(qw.QSizePolicy.Expanding)
        lv.setModel(self.model)
        lv.doubleClicked.connect(self.goto_snapshot)
        lv.setSelectionMode(qw.QAbstractItemView.ExtendedSelection)
        lv.setDragDropMode(qw.QAbstractItemView.InternalMove)
        lv.setEditTriggers(qw.QAbstractItemView.NoEditTriggers)
        lv.viewport().setAcceptDrops(True)
        self.item_delegate = SnapshotItemDelegate(self.model, lv)
        lv.setItemDelegate(self.item_delegate)
        self.list_view = lv
        layout.addWidget(lv, 0, 0, 1, 3)

        pb = qw.QPushButton('New')
        pb.clicked.connect(self.take_snapshot)
        layout.addWidget(pb, 1, 0, 1, 1)

        pb = qw.QPushButton('Replace')
        pb.clicked.connect(self.replace_snapshot)
        layout.addWidget(pb, 1, 1, 1, 1)

        pb = qw.QPushButton('Delete')
        pb.clicked.connect(self.delete_snapshots)
        layout.addWidget(pb, 1, 2, 1, 1)

        self.window_to_image_filter = None

    def setup_menu(self, menu):
        menu.addAction(
            'New',
            self.take_snapshot,
            qg.QKeySequence(qc.Qt.CTRL | qc.Qt.Key_N)).setShortcutContext(
                qc.Qt.ApplicationShortcut)

        menu.addSeparator()

        menu.addAction(
            'Next',
            self.transition_to_next_snapshot,
            qg.QKeySequence(qc.Qt.Key_PageDown)).setShortcutContext(
                qc.Qt.ApplicationShortcut)

        menu.addAction(
            'Previous',
            self.transition_to_previous_snapshot,
            qg.QKeySequence(qc.Qt.Key_PageUp)).setShortcutContext(
                qc.Qt.ApplicationShortcut)

        menu.addSeparator()

        menu.addAction(
            'Import...',
            self.import_snapshots)

        menu.addAction(
            'Export...',
            self.export_snapshots)

        menu.addAction(
            'Animate',
            self.animate_snapshots)

        menu.addAction(
            'Export Movie...',
            self.render_movie)

        menu.addSeparator()

        menu.addAction(
            'Show Panel',
            self.show_and_raise)

    def show_and_raise(self):
        self.viewer.raise_panel(self)

    def get_snapshot_image(self):
        if not self.window_to_image_filter:
            wif = vtk.vtkWindowToImageFilter()
            wif.SetInput(self.viewer.renwin)
            wif.SetInputBufferTypeToRGBA()
            wif.ReadFrontBufferOff()
            self.window_to_image_filter = wif

            writer = vtk.vtkPNGWriter()
            writer.SetInputConnection(wif.GetOutputPort())
            writer.SetWriteToMemory(True)
            self.png_writer = writer

        self.viewer.renwin.Render()
        self.window_to_image_filter.Modified()
        self.png_writer.Write()
        data = vtk_to_numpy(self.png_writer.GetResult()).tobytes()
        img = qg.QImage()
        img.loadFromData(data)
        return img

    def get_snapshot_thumbnail(self):
        return self.get_snapshot_image().scaled(
            thumb_size[0], thumb_size[1],
            qc.Qt.KeepAspectRatio, qc.Qt.SmoothTransformation)

    def get_snapshot_thumbnail_png(self):
        img = self.get_snapshot_thumbnail()

        ba = qc.QByteArray()
        buf = qc.QBuffer(ba)
        buf.open(qc.QIODevice.WriteOnly)
        img.save(buf, format='PNG')
        return ba.data()

    def take_snapshot(self):
        self.model.add_snapshot(
            Snapshot(
                state=clone(self.viewer.state),
                thumb=self.get_snapshot_thumbnail_png()))
        self.viewer.raise_panel(self)

    def replace_snapshot(self):
        state = clone(self.viewer.state)
        selected_indexes = self.list_view.selectedIndexes()

        if len(selected_indexes) == 1:
            self.model.replace_snapshot(
                selected_indexes[0],
                Snapshot(
                    state,
                    thumb=self.get_snapshot_thumbnail_png()))

        self.list_view.update()

    def goto_snapshot(self, index):
        item = self.model.get_item_or_none(index)
        if isinstance(item, Snapshot):
            self.viewer.set_state(item.state)
        elif isinstance(item, Transition):
            snap1 = self.model.get_item_or_none(index.row()-1)
            snap2 = self.model.get_item_or_none(index.row()+1)
            if isinstance(snap1, Snapshot) and isinstance(snap2, Snapshot):
                ip = Interpolator(
                    [0.0, item.effective_duration],
                    [snap1.state, snap2.state])

                self.viewer.start_animation(ip)

    def transition_to_next_snapshot(self, direction=1):
        index = self.list_view.currentIndex()
        if index.row() == -1:
            if direction == 1:
                index = self.model.createIndex(0, 0)

        item = self.model.get_item_or_none(index)
        if item is None:
            return

        if isinstance(item, Snapshot):
            snap1 = item
            transition = self.model.get_item_or_none(index.row()+1*direction)
            snap2 = self.model.get_item_or_none(index.row()+2*direction)
        elif isinstance(item, Transition):
            snap1 = self.model.get_item_or_none(index.row()-1*direction)
            transition = item
            snap2 = self.model.get_item_or_none(index.row()+1*direction)

        if None not in (snap1, transition, snap2):
            ip = Interpolator(
                [0.0, transition.effective_duration],
                [snap1.state, snap2.state])

            index = self.model.get_index_for_item(snap2)
            self.list_view.setCurrentIndex(index)

            self.viewer.start_animation(ip)

        elif snap2 is not None:
            index = self.model.get_index_for_item(snap2)
            self.list_view.setCurrentIndex(index)
            self.viewer.set_state(snap2.state)

    def transition_to_previous_snapshot(self):
        self.transition_to_next_snapshot(-1)

    def delete_snapshots(self):
        selected_indexes = self.list_view.selectedIndexes()
        self.model.remove_snapshots(selected_indexes)

    def animate_snapshots(self, **kwargs):
        selected_indexes = self.list_view.selectedIndexes()
        items = self.model.get_series(selected_indexes)

        time_state = []
        item_previous = None
        t = 0.0
        for i, item in enumerate(items):
            item_next = getitem_or_none(items, i+1)
            item_previous = getitem_or_none(items, i-1)

            if isinstance(item, Snapshot):
                time_state.append((t, item.state))
                if item.effective_duration > 0:
                    time_state.append((t+item.effective_duration, item.state))

                t += item.effective_duration

            elif isinstance(item, Transition):
                if None not in (item_previous, item_next) \
                        and item.effective_duration != 0.0:

                    t += item.effective_duration

            item_previous = item

        if len(time_state) < 2:
            return

        ip = Interpolator(*zip(*time_state))

        self.viewer.start_animation(
            ip, output_path=kwargs.get('output_path', None))

    def render_movie(self):
        try:
            check_call(['ffmpeg', '-loglevel', 'panic'])
        except CalledProcessError:
            pass
        except (TypeError, FileNotFoundError):
            logger.warn(
                'Package ffmpeg needed for movie rendering. Please install it '
                '(e.g. on linux distr. via sudo apt-get ffmpeg.) and retry.')
            return

        caption = 'Export Movie'
        fn_out, _ = qw.QFileDialog.getSaveFileName(
            self, caption, 'movie.mp4',
            options=common.qfiledialog_options)

        if fn_out:
            self.animate_snapshots(output_path=fn_out)

    def export_snapshots(self):
        caption = 'Export Snapshots'
        fn, _ = qw.QFileDialog.getSaveFileName(
            self, caption, options=common.qfiledialog_options)

        selected_indexes = self.list_view.selectedIndexes()
        items = self.model.get_series(selected_indexes)

        if fn:
            dump_all(items, filename=fn)

    def add_snapshots(self, snapshots):
        self.model.append_series(snapshots)

    def load_snapshots(self, path):
        items = load_snapshots(path)
        self.add_snapshots(items)

    def import_snapshots(self):
        caption = 'Import Snapshots'
        path, _ = qw.QFileDialog.getOpenFileName(
            self, caption, options=common.qfiledialog_options)

        if path:
            self.load_snapshots(path)


class Item(Object):
    duration = Float.T(optional=True)

    def __init__(self, **kwargs):
        Object.__init__(self, **kwargs)
        self.auto_duration = 0.0

    @property
    def effective_duration(self):
        if self.duration is not None:
            return self.duration
        else:
            return self.auto_duration


class Snapshot(Item):
    name = String.T()
    state = ViewerState.T()
    thumb = Bytes.T(optional=True)

    isnapshot = 0

    def __init__(self, state, name=None, thumb=None, **kwargs):

        if name is None:
            Snapshot.isnapshot += 1
            name = '%i' % Snapshot.isnapshot

        Item.__init__(self, state=state, name=name, thumb=thumb, **kwargs)
        self._img = None

    def get_name(self):
        return self.name

    def get_image(self):
        if self.thumb is not None and not self._img:
            img = qg.QImage()
            img.loadFromData(self.thumb)
            self._img = img

        return self._img


class Transition(Item):

    def __init__(self, **kwargs):
        Item.__init__(self, **kwargs)
        self.animate = []

    def get_name(self):
        ed = self.effective_duration
        return '%s %s' % (
            'T' if self.animate and self.effective_duration > 0.0 else '',
            '%.2f s' % ed if ed != 0.0 else '')

    @property
    def name(self):
        return self.get_name()


class SnapshotsModel(qc.QAbstractListModel):

    def __init__(self):
        qc.QAbstractListModel.__init__(self)
        self._items = []

    def supportedDropActions(self):
        return qc.Qt.MoveAction

    def rowCount(self, parent=None):
        return len(self._items)

    def insertRows(self, index):
        pass

    def mimeTypes(self):
        return ['text/plain']

    def mimeData(self, indices):
        objects = [self._items[i.row()] for i in indices]
        serialized = dump_all(objects)
        md = qc.QMimeData()
        md.setText(serialized)
        md._item_objects = objects
        return md

    def dropMimeData(self, md, action, row, col, index):
        i = index.row()
        items = getattr(md, '_item_objects', [])
        self.beginInsertRows(qc.QModelIndex(), i, i)
        self._items[i:i] = items
        self.endInsertRows()
        n = len(items)
        joff = 0
        for j in range(len(self._items)):
            if (j < i or j >= i+n) and self._items[j+joff] in items:
                self.beginRemoveRows(qc.QModelIndex(), j+joff, j+joff)
                self._items[j+joff:j+joff+1] = []
                self.endRemoveRows()
                joff -= 1

        self.repair_transitions()
        return True

    def removeRows(self, i, n, parent):
        return True

    def flags(self, index):
        if index.isValid():
            i = index.row()
            if isinstance(self._items[i], Snapshot):
                return qc.Qt.ItemFlags(
                    qc.Qt.ItemIsSelectable
                    | qc.Qt.ItemIsEnabled
                    | qc.Qt.ItemIsDragEnabled
                    | qc.Qt.ItemIsEditable)

            else:
                return qc.Qt.ItemFlags(
                    qc.Qt.ItemIsEnabled
                    | qc.Qt.ItemIsEnabled
                    | qc.Qt.ItemIsDropEnabled
                    | qc.Qt.ItemIsEditable)
        else:
            return qc.QAbstractListModel.flags(self, index)

    def data(self, index, role):
        app = common.get_app()
        i = index.row()
        item = self._items[i]
        is_snap = isinstance(item, Snapshot)
        if role == qc.Qt.DisplayRole:
            if is_snap:
                return qc.QVariant(str(item.get_name()))
            else:
                return qc.QVariant(str(item.get_name()))

        elif role == qc.Qt.ToolTipRole:
            if is_snap:
                # return qc.QVariant(str(item.state))
                return qc.QVariant()
            else:
                if item.animate:
                    label = 'Interpolation: %s' % \
                        ', '.join(x[0] for x in item.animate)
                else:
                    label = 'Not interpolable.'

                return qc.QVariant(label)

        elif role == qc.Qt.TextAlignmentRole and not is_snap:
            return qc.QVariant(qc.Qt.AlignRight)

        elif role == qc.Qt.ForegroundRole and not is_snap:
            if item.duration is None:
                return qc.QVariant(app.palette().brush(
                    qg.QPalette.Disabled, qg.QPalette.Text))
            else:
                return qc.QVariant(app.palette().brush(
                    qg.QPalette.Active, qg.QPalette.Text))

        else:
            qc.QVariant()

    def headerData(self):
        pass

    def add_snapshot(self, snapshot):
        self.beginInsertRows(
            qc.QModelIndex(), self.rowCount(), self.rowCount())
        self._items.append(snapshot)
        self.endInsertRows()
        self.repair_transitions()

    def replace_snapshot(self, index, snapshot):
        self._items[index.row()] = snapshot
        self.dataChanged.emit(index, index)
        self.repair_transitions()

    def remove_snapshots(self, indexes):
        indexes = sorted(indexes, key=lambda index: index.row())
        ioff = 0
        for index in indexes:
            i = index.row()
            self.beginRemoveRows(qc.QModelIndex(), i+ioff, i+ioff)
            self._items[i+ioff:i+ioff+1] = []
            self.endRemoveRows()
            ioff -= 1

        self.repair_transitions()

    def repair_transitions(self):
        items = self._items
        i = 0
        need = 0
        while i < len(items):
            if need == 0:
                if not isinstance(items[i], Transition):
                    self.beginInsertRows(qc.QModelIndex(), i, i)
                    items[i:i] = [Transition()]
                    self.endInsertRows()
                else:
                    i += 1
                    need = 1
            elif need == 1:
                if not isinstance(items[i], Snapshot):
                    self.beginRemoveRows(qc.QModelIndex(), i, i)
                    items[i:i+1] = []
                    self.endRemoveRows()
                else:
                    i += 1
                    need = 0

        if len(items) == 1:
            self.beginRemoveRows(qc.QModelIndex(), 0, 0)
            items[:] = []
            self.endRemoveRows()

        elif len(items) > 1:
            if not isinstance(items[-1], Transition):
                self.beginInsertRows(
                    qc.QModelIndex(), self.rowCount(), self.rowCount())
                items.append(Transition())
                self.endInsertRows()

        self.update_auto_durations()

    def update_auto_durations(self):
        items = self._items
        for i, item in enumerate(items):
            if isinstance(item, Transition):
                if 0 < i < len(items)-1:
                    item.animate = interpolateables(
                        items[i-1].state, items[i+1].state)

                    if item.animate:
                        item.auto_duration = 1.
                    else:
                        item.auto_duration = 0.

        for i, item in enumerate(items):
            if isinstance(item, Snapshot):
                if 0 < i < len(items)-1:
                    if items[i-1].effective_duration == 0 \
                            and items[i+1].effective_duration == 0:
                        item.auto_duration = 1.
                    else:
                        item.auto_duration = 0.

    def get_index_for_item(self, item):
        for i, candidate in enumerate(self._items):
            if candidate is item:
                return self.createIndex(i, 0)

        return None

    def get_item_or_none(self, index):
        if not isinstance(index, int):
            i = index.row()
        else:
            i = index

        if i < 0 or len(self._items) <= i:
            return None

        try:
            return self._items[i]
        except IndexError:
            return None

    def get_series(self, indexes):
        items = self._items

        ilist = sorted([index.row() for index in indexes])
        if len(ilist) <= 1:
            ilist = list(range(0, len(self._items)))

        ilist = [i for i in ilist if isinstance(items[i], Snapshot)]
        if len(ilist) == 0:
            return []

        i = ilist[0]

        series = []
        while ilist:
            i = ilist.pop(0)
            series.append(items[i])
            if ilist and ilist[0] == i+2:
                series.append(items[i+1])

        return series

    def append_series(self, items):
        self.beginInsertRows(
            qc.QModelIndex(),
            self.rowCount(), self.rowCount() + len(items) - 1)

        self._items.extend(items)
        self.endInsertRows()

        self.repair_transitions()


def load_snapshots(path):

    if path.startswith('http://') or path.startswith('https://'):
        with tempfile.NamedTemporaryFile() as fp:
            util.download_file(path, fp.name)
            return load_snapshots(fp.name)

    items = load_all(filename=path)
    for i in range(len(items)):
        if not isinstance(
                items[i], (ViewerState, Snapshot, Transition)):

            logger.warn(
                'Only Snapshot, Transition and ViewerState objects '
                'are accepted. Object #%i from file %s ignored.'
                % (i, path))

        if isinstance(items[i], ViewerState):
            items[i] = Snapshot(items[i])

    for item in items:
        if isinstance(item, Snapshot):
            item.state.sort_elements()

    return items



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from . import base
from .base import *  # noqa

__doc__ = base.__doc__

__all__ = base.__all__



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import sys
import time
import logging

from .get_terminal_size import get_terminal_size

logger = logging.getLogger('pyrockoeost.progress')

# TODO: Refactor so that if multiple viewers are attached, they can do
# their updates independently of each other (at independent time intervals).

# TODO: Refactor so that different viewers can render task states differently.

# spinner = u'\u25dc\u25dd\u25de\u25df'
# spinner = '⣾⣽⣻⢿⡿⣟⣯⣷'
spinner = '◴◷◶◵'
skull = u'\u2620'
check = u'\u2714'
cross = u'\u2716'
bar = u'[- ]'
blocks = u'\u2588\u2589\u258a\u258b\u258c\u258d\u258e\u258f '

symbol_done = check
symbol_failed = cross  # skull

ansi_up = u'\033[%iA'
ansi_down = u'\033[%iB'
ansi_left = u'\033[%iC'
ansi_right = u'\033[%iD'
ansi_next_line = u'\033E'

ansi_erase_display = u'\033[2J'
ansi_window = u'\033[%i;%ir'
ansi_move_to = u'\033[%i;%iH'

ansi_clear_down = u'\033[0J'
ansi_clear_up = u'\033[1J'
ansi_clear = u'\033[2J'

ansi_clear_right = u'\033[0K'

ansi_scroll_up = u'\033D'
ansi_scroll_down = u'\033M'

ansi_reset = u'\033c'


g_force_viewer_off = False

g_viewer = 'terminal'


def set_default_viewer(viewer):
    global g_viewer
    g_viewer = viewer


class StatusViewer(object):

    def __init__(self, parent=None):
        self._parent = parent

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.stop()

    def stop(self):
        if self._parent:
            self._parent.hide(self)

    def draw(self, lines):
        pass


class TerminalStatusViewer(StatusViewer):
    def __init__(self, parent=None):
        self._terminal_size = get_terminal_size()
        self._height = 0
        self._state = 0
        self._parent = parent

    def print(self, s):
        print(s, end='', file=sys.stderr)

    def flush(self):
        print('', end='', flush=True, file=sys.stderr)

    def start(self):
        sx, sy = self._terminal_size
        self._state = 1

    def stop(self):
        if self._state == 1:
            sx, sy = self._terminal_size
            self._resize(0)
            self.print(ansi_move_to % (sy-self._height, 1))
            self.flush()

        self._state = 2
        if self._parent:
            self._parent.hide(self)

    def _start_show(self):
        sx, sy = self._terminal_size
        self.print(ansi_move_to % (sy-self._height+1, 1))

    def _end_show(self):
        sx, sy = self._terminal_size
        self.print(ansi_move_to % (sy-self._height, 1))
        self.print(ansi_clear_right)

    def _resize(self, height):
        sx, sy = self._terminal_size
        k = height - self._height
        if k > 0:
            self.print(ansi_scroll_up * k)
            self.print(ansi_window % (1, sy-height))
        if k < 0:
            self.print(ansi_window % (1, sy-height))
            self.print(ansi_scroll_down * abs(k))

        self._height = height

    def draw(self, lines):
        if self._state == 0:
            self.start()

        if self._state != 1:
            return

        self._terminal_size = get_terminal_size()
        sx, sy = self._terminal_size
        nlines = len(lines)
        self._resize(nlines)
        self._start_show()

        for iline, line in enumerate(reversed(lines)):
            if len(line) > sx - 1:
                line = line[:sx-1]

            self.print(ansi_clear_right + line)
            if iline != nlines - 1:
                self.print(ansi_next_line)

        self._end_show()
        self.flush()


class LogStatusViewer(StatusViewer):

    def draw(self, lines):
        if lines:
            logger.info(
                'Progress:\n%s' % '\n'.join('  '+line for line in lines))


class DummyStatusViewer(StatusViewer):
    pass


class Task(object):
    def __init__(
            self, progress, id, name, n, state='working', logger=None,
            group=None):

        self._id = id
        self._name = name
        self._condition = ''
        self._ispin = 0
        self._i = None
        self._n = n
        self._done = False
        assert state in ('waiting', 'working')
        self._state = state
        self._progress = progress
        self._logger = logger
        self._tcreate = time.time()
        self._group = group

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        if type is None:
            self.done()
        else:
            self.fail()

    def __call__(self, it):
        try:
            self._n = len(it)
        except TypeError:
            self._n = None

        clean = False
        try:
            n = 0
            for obj in it:
                self.update(n)
                yield obj
                n += 1

            self.update(n)
            clean = True

        finally:
            if clean:
                self.done()
            else:
                self.fail()

    def log(self, s):
        if self._logger is not None:
            self._logger.info(s)

    def get_group_time_start(self):
        if self._group:
            return self._group.get_group_time_start()
        else:
            return self._tcreate

    def task(self, *args, **kwargs):
        kwargs['group'] = self
        return self._progress.task(*args, **kwargs)

    def update(self, i=None, condition=''):
        self._state = 'working'

        self._condition = condition

        if i is not None:
            if self._n is not None:
                i = min(i, self._n)

            self._i = i

        self._progress._update()

    def done(self, condition=''):
        self.duration = time.time() - self._tcreate

        if self._state in ('done', 'failed'):
            return

        self._condition = condition
        self._state = 'done'
        self._progress._end(self)
        self.log(str(self))

    def fail(self, condition=''):
        self.duration = time.time() - self._tcreate

        if self._state in ('done', 'failed'):
            return

        self._condition = condition
        self._state = 'failed'
        self._progress._end(self)
        self.log(str(self))

    def _str_state(self):
        s = self._state
        if s == 'waiting':
            return '  '
        elif s == 'working':
            self._ispin += 1
            return spinner[self._ispin % len(spinner)] + ' '
        elif s == 'done':
            return symbol_done + ' '
        elif s == 'failed':
            return symbol_failed + ' '
        else:
            return '? '

    def _idisplay(self):
        i = self._i
        if self._n is not None and i > self._n:
            i = self._n
        return i

    def _str_progress(self):
        if self._i is None:
            return self._state
        elif self._n is None:
            if self._state != 'working':
                return '... %s (%i)' % (self._state, self._idisplay())
            else:
                return '%i' % self._idisplay()
        else:
            if self._state == 'working':
                nw = len(str(self._n))
                return (('%' + str(nw) + 'i / %i') % (
                    self._idisplay(), self._n)).center(11)

            elif self._state == 'failed':
                return '... %s (%i / %i)' % (
                    self._state, self._idisplay(), self._n)
            else:
                return '... %s (%i)' % (self._state, self._n)

    def _str_percent(self):
        if self._state == 'working' and self._n is not None and self._n >= 4 \
                and self._i is not None:
            return '%3.0f%%' % ((100. * self._i) / self._n)
        else:
            return ''

    def _str_condition(self):
        if self._condition:
            return '%s' % self._condition
        else:
            return ''

    def _str_bar(self):
        if self._state == 'working' and self._n is not None and self._n >= 4 \
                and self._i is not None:
            nb = 20
            fb = nb * float(self._i) / self._n
            ib = int(fb)
            ip = int((fb - ib) * (len(blocks)-1))
            if ib == 0 and ip == 0:
                ip = 1  # indication of start
            s = blocks[0] * ib
            if ib < nb:
                s += blocks[-1-ip] + (nb - ib - 1) * blocks[-1] + blocks[-2]

            # s = ' ' + bar[0] + bar[1] * ib + bar[2] * (nb - ib) + bar[3]
            return s
        else:
            return ''

    def __str__(self):
        return '%s%-23s %-11s %s%-4s  %s' % (
            self._str_state(),
            self._name,
            self._str_progress(),
            self._str_bar(),
            self._str_percent(),
            self._str_condition())


class Progress(object):

    def __init__(self):
        self._current_id = 0
        self._current_group_id = 0
        self._tasks = {}
        self._tasks_done = []
        self._last_update = 0.0
        self._terms = []

    def view(self, viewer=None):
        if g_force_viewer_off or self._terms:
            viewer = 'off'
        elif viewer is None:
            viewer = g_viewer

        try:
            term = g_viewer_classes[viewer](self)
        except KeyError:
            raise ValueError('Invalid viewer choice: %s' % viewer)

        self._terms.append(term)
        return term

    def hide(self, term):
        self._update(force=True)
        self._terms.remove(term)

    def task(self, name, n=None, logger=None, group=None):
        self._current_id += 1
        task = Task(
            self, self._current_id, name, n, logger=logger, group=group)
        self._tasks[task._id] = task
        self._update(force=True)
        return task

    def _end(self, task):
        del self._tasks[task._id]
        self._tasks_done.append(task)
        self._update(force=True)

    def _update(self, force=False):
        now = time.time()
        if self._last_update + 0.1 < now or force:
            self._tasks_done = []

            lines = self._lines()
            for term in self._terms:
                term.draw(lines)

            self._last_update = now

    def _lines(self):
        task_ids = sorted(self._tasks)
        lines = []
        for task_id in task_ids:
            task = self._tasks[task_id]
            lines.extend(str(task).splitlines())

        return lines


g_viewer_classes = {
    'terminal': TerminalStatusViewer,
    'log': LogStatusViewer,
    'off': DummyStatusViewer}

progress = Progress()



# http://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

import sys

from pyrockoeost.squirrel.check import do_check, SquirrelCheckProblemType

from ..common import csvtype, dq, ldq

headline = 'Check dataset consistency.'


def make_subparser(subparsers):
    return subparsers.add_parser(
        'check',
        help=headline,
        description=headline + '''

A report listing potential dataset/metadata problems for a given data
collection is printed to standard output. The following problems are detected:

%s
''' % '\n'.join(
            '  [%s]: %s' % (k, v)
            for (k, v) in SquirrelCheckProblemType.types.items()))


def setup(parser):
    parser.add_squirrel_selection_arguments()
    parser.add_squirrel_query_arguments(without=['kinds'])
    parser.add_argument(
        '--ignore',
        type=csvtype(SquirrelCheckProblemType.choices),
        metavar='TYPE,...',
        help='Problem types to be ignored. Choices: %s.'
        % ldq(SquirrelCheckProblemType.choices))

    parser.add_argument(
        '--verbose', '-v',
        action='count',
        dest='verbosity',
        default=0,
        help='Verbose mode for textual output. Multiple ``-v`` options '
             'increase the verbosity. The maximum is 2. At level 1, ``ok`` '
             'indicators are printed for entries with no problem. At level 2, '
             'availability information is shown for each entry.')

    format_default = 'text'
    format_choices = ['text', 'yaml']

    parser.add_argument(
        '--out-format',
        choices=format_choices,
        default=format_default,
        dest='output_format',
        metavar='FMT',
        help='Specifies how output is presented. Choices: %s. '
             'Default: %s.' % (ldq(format_choices), dq(format_default)))


def run(parser, args):
    squirrel = args.make_squirrel()

    check = do_check(
        squirrel,
        ignore=args.ignore or [],
        **args.squirrel_query)

    if args.output_format == 'text':
        print(check.get_text(verbosity=args.verbosity))
    elif args.output_format == 'yaml':
        print(check)
    else:
        assert False

    sys.exit(0 if check.get_nproblems() == 0 else 1)



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

'''
Effective seismological trace viewer.
'''

import sys
import logging
import time
import re
import zlib
import struct
import pickle
try:
    from urlparse import parse_qsl
except ImportError:
    from urllib.parse import parse_qsl

# EOST
from PyQt5.QtGui import QPixmap
from pyrockoeost.gui.snuffling import MyScrollArea

from pyrockoeost.streaming import serial_hamster
from pyrockoeost.streaming import slink
from pyrockoeost.streaming import edl
from pyrockoeost.streaming import datacube

from pyrockoeost import pile            # noqa
from pyrockoeost import util            # noqa
from pyrockoeost import model           # noqa
from pyrockoeost import config          # noqa
from pyrockoeost import io              # noqa

from . import pile_viewer     # noqa

from ..qt_compat import qc, qg, qw

logger = logging.getLogger('pyrockoeost.gui.snuffler.snuffler_app')


class AcquisitionThread(qc.QThread):
    def __init__(self, post_process_sleep=0.0):
        qc.QThread.__init__(self)
        self.mutex = qc.QMutex()
        self.queue = []
        self.post_process_sleep = post_process_sleep
        self._sun_is_shining = True

    def get_wanted_poll_interval(self):
        return 1000.

    def run(self):
        while True:
            try:
                self.acquisition_start()
                while self._sun_is_shining:
                    t0 = time.time()
                    self.process()
                    t1 = time.time()
                    if self.post_process_sleep != 0.0:
                        time.sleep(max(0, self.post_process_sleep-(t1-t0)))

                self.acquisition_stop()
                break

            except (
                    edl.ReadError,
                    serial_hamster.SerialHamsterError,
                    slink.SlowSlinkError) as e:

                logger.error(str(e))
                logger.error('Acquistion terminated, restart in 5 s')
                self.acquisition_stop()
                time.sleep(5)
                if not self._sun_is_shining:
                    break

    def stop(self):
        self._sun_is_shining = False

        logger.debug('Waiting for thread to terminate...')
        self.wait()
        logger.debug('Thread has terminated.')

    def got_trace(self, tr):
        self.mutex.lock()
        self.queue.append(tr)
        self.mutex.unlock()

    def poll(self):
        self.mutex.lock()
        items = self.queue[:]
        self.queue[:] = []
        self.mutex.unlock()
        return items


class SlinkAcquisition(
        slink.SlowSlink, AcquisitionThread):

    def __init__(self, *args, **kwargs):
        slink.SlowSlink.__init__(self, *args, **kwargs)
        AcquisitionThread.__init__(self)

    def got_trace(self, tr):
        AcquisitionThread.got_trace(self, tr)


class CamAcquisition(
        serial_hamster.CamSerialHamster, AcquisitionThread):

    def __init__(self, *args, **kwargs):
        serial_hamster.CamSerialHamster.__init__(self, *args, **kwargs)
        AcquisitionThread.__init__(self, post_process_sleep=0.1)

    def got_trace(self, tr):
        AcquisitionThread.got_trace(self, tr)


class USBHB628Acquisition(
        serial_hamster.USBHB628Hamster, AcquisitionThread):

    def __init__(self, deltat=0.02, *args, **kwargs):
        serial_hamster.USBHB628Hamster.__init__(
            self, deltat=deltat, *args, **kwargs)
        AcquisitionThread.__init__(self)

    def got_trace(self, tr):
        AcquisitionThread.got_trace(self, tr)


class SchoolSeismometerAcquisition(
        serial_hamster.SerialHamster, AcquisitionThread):

    def __init__(self, *args, **kwargs):
        serial_hamster.SerialHamster.__init__(self, *args, **kwargs)
        AcquisitionThread.__init__(self, post_process_sleep=0.0)

    def got_trace(self, tr):
        AcquisitionThread.got_trace(self, tr)

    def get_wanted_poll_interval(self):
        return 100.


class EDLAcquisition(
        edl.EDLHamster, AcquisitionThread):

    def __init__(self, *args, **kwargs):
        edl.EDLHamster.__init__(self, *args, **kwargs)
        AcquisitionThread.__init__(self)

    def got_trace(self, tr):
        AcquisitionThread.got_trace(self, tr)


class CubeAcquisition(
        datacube.SerialCube, AcquisitionThread):

    def __init__(self, *args, **kwargs):
        datacube.SerialCube.__init__(self, *args, **kwargs)
        AcquisitionThread.__init__(self)

    def got_trace(self, tr):
        AcquisitionThread.got_trace(self, tr)


def setup_acquisition_sources(args):

    sources = []
    iarg = 0
    while iarg < len(args):
        arg = args[iarg]

        msl = re.match(r'seedlink://([a-zA-Z0-9.-]+)(:(\d+))?(/(.*))?', arg)
        mca = re.match(r'cam://([^:]+)', arg)
        mus = re.match(r'hb628://([^:?]+)(\?([^?]+))?', arg)
        msc = re.match(r'school://([^:?]+)(\?([^?]+))?', arg)
        med = re.match(r'edl://([^:]+)', arg)
        mcu = re.match(r'cube://([^:]+)', arg)

        if msl:
            host = msl.group(1)
            port = msl.group(3)
            if not port:
                port = '18000'

            sl = SlinkAcquisition(host=host, port=port)
            if msl.group(5):
                stream_patterns = msl.group(5).split(',')

                if '_' not in msl.group(5):
                    try:
                        streams = sl.query_streams()
                    except slink.SlowSlinkError as e:
                        logger.fatal(str(e))
                        sys.exit(1)

                    streams = list(set(
                        util.match_nslcs(stream_patterns, streams)))

                    for stream in streams:
                        sl.add_stream(*stream)
                else:
                    for stream in stream_patterns:
                        sl.add_raw_stream_selector(stream)

            sources.append(sl)

        elif mca:
            port = mca.group(1)
            cam = CamAcquisition(port=port, deltat=0.0314504)
            sources.append(cam)

        elif mus:
            port = mus.group(1)
            try:
                d = {}
                if mus.group(3):
                    d = dict(parse_qsl(mus.group(3)))

                deltat = 1.0/float(d.get('rate', '50'))
                channels = [(int(c), c) for c in d.get('channels', '01234567')]
                hb628 = USBHB628Acquisition(
                    port=port,
                    deltat=deltat,
                    channels=channels,
                    buffersize=16,
                    lookback=50)

                sources.append(hb628)
            except Exception:
                sys.exit('invalid acquisition source: %s' % arg)

        elif msc:
            port = msc.group(1)

            d = {}
            if msc.group(3):
                d = dict(parse_qsl(msc.group(3)))

            d_rate = {
                '20': ('a', 20.032),
                '40': ('b', 39.860),
                '80': ('c', 79.719)}

            s_rate = d.get('rate', '80')
            station = d.get('station', 'TEST')

            if s_rate not in d_rate:
                raise Exception(
                    'Unsupported rate: %s (expected "20", "40" or "80")'
                    % s_rate)

            s_gain = d.get('gain', '4')

            if s_gain not in ('1', '2', '4'):
                raise Exception(
                    'Unsupported gain: %s (expected "1", "2" or "4")'
                    % s_gain)

            start_string = s_gain + d_rate[s_rate][0]
            deltat = 1.0 / d_rate[s_rate][1]

            logger.info(
                'School seismometer: trying to use device %s with gain=%s and '
                'rate=%g.' % (port, s_gain, 1.0/deltat))

            sco = SchoolSeismometerAcquisition(
                port=port,
                deltat=deltat,
                start_string=start_string,
                min_detection_size=50,
                disallow_uneven_sampling_rates=False,
                station=station)

            sources.append(sco)

        elif med:
            port = med.group(1)
            edl = EDLAcquisition(port=port)
            sources.append(edl)
        elif mcu:
            device = mcu.group(1)
            cube = CubeAcquisition(device=device)
            sources.append(cube)

        if msl or mca or mus or msc or med or mcu:
            args.pop(iarg)
        else:
            iarg += 1

    return sources


class PollInjector(qc.QObject):

    def __init__(self, *args, **kwargs):
        interval = kwargs.pop('interval', 1000.)
        qc.QObject.__init__(self)
        self._injector = pile.Injector(*args, **kwargs)
        self._sources = []
        self.startTimer(int(interval))

    def add_source(self, source):
        self._sources.append(source)

    def remove_source(self, source):
        self._sources.remove(source)

    def timerEvent(self, ev):
        for source in self._sources:
            trs = source.poll()
            for tr in trs:
                self._injector.inject(tr)

    # following methods needed because mulitple inheritance does not seem
    # to work anymore with QObject in Python3 or PyQt5

    def set_fixation_length(self, length):
        return self._injector.set_fixation_length(length)

    def set_save_path(
            self,
            path='dump_%(network)s.%(station)s.%(location)s.%(channel)s_'
                 '%(tmin)s_%(tmax)s.mseed'):

        return self._injector.set_save_path(path)

    def fixate_all(self):
        return self._injector.fixate_all()

    def free(self):
        return self._injector.free()


class Connection(qc.QObject):

    received = qc.pyqtSignal(object, object)
    disconnected = qc.pyqtSignal(object)

    def __init__(self, parent, sock):
        qc.QObject.__init__(self, parent)
        self.socket = sock
        self.readyRead.connect(
            self.handle_read)
        self.disconnected.connect(
            self.handle_disconnected)
        self.nwanted = 8
        self.reading_size = True
        self.handler = None
        self.nbytes_received = 0
        self.nbytes_sent = 0
        self.compressor = zlib.compressobj()
        self.decompressor = zlib.decompressobj()

    def handle_read(self):
        while True:
            navail = self.socket.bytesAvailable()
            if navail < self.nwanted:
                return

            data = self.socket.read(self.nwanted)
            self.nbytes_received += len(data)
            if self.reading_size:
                self.nwanted = struct.unpack('>Q', data)[0]
                self.reading_size = False
            else:
                obj = pickle.loads(self.decompressor.decompress(data))
                if obj is None:
                    self.socket.disconnectFromHost()
                else:
                    self.handle_received(obj)
                self.nwanted = 8
                self.reading_size = True

    def handle_received(self, obj):
        self.received.emit(self, obj)

    def ship(self, obj):
        data = self.compressor.compress(pickle.dumps(obj))
        data_end = self.compressor.flush(zlib.Z_FULL_FLUSH)
        self.socket.write(struct.pack('>Q', len(data)+len(data_end)))
        self.socket.write(data)
        self.socket.write(data_end)
        self.nbytes_sent += len(data)+len(data_end) + 8

    def handle_disconnected(self):
        self.disconnected.emit(self)

    def close(self):
        self.socket.close()


class ConnectionHandler(qc.QObject):
    def __init__(self, parent):
        qc.QObject.__init__(self, parent)
        self.queue = []
        self.connection = None

    def connected(self):
        return self.connection is None

    def set_connection(self, connection):
        self.connection = connection
        connection.received.connect(
            self._handle_received)

        connection.connect(
            self.handle_disconnected)

        for obj in self.queue:
            self.connection.ship(obj)

        self.queue = []

    def _handle_received(self, conn, obj):
        self.handle_received(obj)

    def handle_received(self, obj):
        pass

    def handle_disconnected(self):
        self.connection = None

    def ship(self, obj):
        if self.connection:
            self.connection.ship(obj)
        else:
            self.queue.append(obj)


class SimpleConnectionHandler(ConnectionHandler):
    def __init__(self, parent, **mapping):
        ConnectionHandler.__init__(self, parent)
        self.mapping = mapping

    def handle_received(self, obj):
        command = obj[0]
        args = obj[1:]
        self.mapping[command](*args)


class MyMainWindow(qw.QMainWindow):

    def __init__(self, app, *args):
        qg.QMainWindow.__init__(self, *args)
        self.app = app

    def keyPressEvent(self, ev):
        self.app.pile_viewer.get_view().keyPressEvent(ev)


class SnufflerTabs(qw.QTabWidget):
    def __init__(self, parent):
        qw.QTabWidget.__init__(self, parent)
        if hasattr(self, 'setTabsClosable'):
            self.setTabsClosable(True)

        self.tabCloseRequested.connect(
            self.removeTab)

        if hasattr(self, 'setDocumentMode'):
            self.setDocumentMode(True)

    def hide_close_button_on_first_tab(self):
        tbar = self.tabBar()
        if hasattr(tbar, 'setTabButton'):
            tbar.setTabButton(0, qw.QTabBar.LeftSide, None)
            tbar.setTabButton(0, qw.QTabBar.RightSide, None)

    def append_tab(self, widget, name):
        widget.setParent(self)
        self.insertTab(self.count(), widget, name)
        self.setCurrentIndex(self.count()-1)

    def remove_tab(self, widget):
        self.removeTab(self.indexOf(widget))

    def tabInserted(self, index):
        if index == 0:
            self.hide_close_button_on_first_tab()

        self.tabbar_visibility()
        self.setFocus()

    def removeTab(self, index):
        w = self.widget(index)
        w.close()
        qw.QTabWidget.removeTab(self, index)

    def tabRemoved(self, index):
        self.tabbar_visibility()

    def tabbar_visibility(self):
        if self.count() <= 1:
            self.tabBar().hide()
        elif self.count() > 1:
            self.tabBar().show()

    def keyPressEvent(self, event):
        if event.text() == 'd':
            i = self.currentIndex()
            if i != 0:
                self.tabCloseRequested.emit(i)
        else:
            self.parent().keyPressEvent(event)


class SnufflerStartWizard(qw.QWizard):

    def __init__(self, parent):
        qw.QWizard.__init__(self, parent)

        self.setOption(self.NoBackButtonOnStartPage)
        self.setOption(self.NoBackButtonOnLastPage)
        self.setOption(self.NoCancelButton)
        self.addPageSurvey()
        self.addPageHelp()
        self.setWindowTitle('Bienvenue dans pyrockoeost')

    def getSystemInfo(self):
        import numpy
        import scipy
        import pyrockoeost
        import platform
        import uuid
        data = {
            'node-uuid': uuid.getnode(),
            'platform.architecture': platform.architecture(),
            'platform.system': platform.system(),
            'platform.release': platform.release(),
            'python': platform.python_version(),
            'pyrockoeost': pyrockoeost.__version__,
            'numpy': numpy.__version__,
            'scipy': scipy.__version__,
            'qt': qc.PYQT_VERSION_STR,
        }
        return data

    def addPageSurvey(self):
        import pprint
        webtk = 'DSFGK234ADF4ASDF'
        sys_info = self.getSystemInfo()

        p = qw.QWizardPage()
        p.setCommitPage(True)
        p.setTitle("Merci d'avoir installé pyrockoeost | <b>EOST version</b> !")

        # lyt = qw.QVBoxLayout()
        # lyt.addWidget(qw.QLabel(
        #     '<p>Your feedback is important for'
        #     ' the development and improvement of pyrockoeost.</p>'
        #     '<p>Do you want to send this system information anon'
        #     'ymously to <a href="https://pyrocko.org">'
        #     'https://pyrocko.org</a>?</p>'))

        lyt = qw.QVBoxLayout()
        lyt.addWidget(qw.QLabel(
            "<p>Vos commentaires sont importants pour"
            " le développement et l'amélioration de pyrockoeost.</p>"
            "<p><b>N'hésitez pas à nous en faire part !</b></p>"))

        text_data = qw.QLabel(
            '<code style="font-size: small;">Vos infos :<br>%s</code>' %
            pprint.pformat(
                sys_info,
                indent=1).replace('\n', '<br>')
            )
        text_data.setStyleSheet('padding: 10px;')
        lyt.addWidget(text_data)

        lyt.addWidget(qw.QLabel(
            "Vous pouvez jeter un œil à la configuration (touches, emplacement)\n dans : ~/.pyrockoeost/snuffler.pf\n\n"
            "Ce message ne sera plus affiché.\n"
            '- The pyrockoeost Developers'
            ))

        p.setLayout(lyt)
        p.setButtonText(self.CommitButton, 'Continuer')

        # yes_btn = qw.QPushButton(p)
        # yes_btn.setText('Yes')

        @qc.pyqtSlot()
        def send_data():
            import requests
            import json
            try:
                requests.post('https://pyrocko.org/%s' % webtk,
                              data=json.dumps(sys_info))
            except Exception as e:
                print(e)
            self.button(self.NextButton).clicked.emit(True)

        # self.customButtonClicked.connect(send_data)

        # self.setButton(self.CustomButton1, yes_btn)
        # self.setOption(self.HaveCustomButton1, True)

        self.addPage(p)
        return p

    def addPageHelp(self):
        p = qw.QWizardPage()
        p.setTitle('Bienvenue dans Snuffler !')

        text = qw.QLabel('''<html>
<h3>- <i>Le navigateur de sismogrammes et son atelier.</i></h3>
<p>On dirait que vous démarrez Snuffler pour la première fois.<br>
Il vous permet de parcourir et de traiter de grandes archives de données sous formes d'ondes.</p>
<p>Le traitement de base est complété par des Snufflings (plugins) comme celui de l'EOST :</p>
<ul>
    <li><b>Télécharger des sismogrammes</b> depuis GEOFON IRIS et d'autres sources</li>
    <li><b>Catalogue de tremblements de terre</b> accès à GEOFON, GlobalCMT, USGS...</li>
    <li><b>Cake</b>, calculer les temps d'arrivée synthétiques</li>
    <li><b>Seismosizer</b>, générer des sismogrammes synthétiques à la volée</li>
    <li>
        <b>Carte</b>, inspecter rapidement les stations et les événements sur des cartes interactives
    </li>
</ul>
<p>Et plus encore, voir <a href="https://pyrocko.org/">https://pyrocko.org/</a></p>

<p style="width: 100%; background-color: #0071b6; margin: 5px; padding: 50;"
          align="center">
    <b>Bonne chance !</b>
</p>
</html>''')

        lyt = qw.QVBoxLayout()
        lyt.addWidget(text)

        def remove_custom_button():
            self.setOption(self.HaveCustomButton1, False)

        p.initializePage = remove_custom_button

        p.setLayout(lyt)
        self.addPage(p)
        p.setButtonText(qw.QWizard.FinishButton, "Terminer")
        return p


class SnufflerWindow(qw.QMainWindow):

    def __init__(
            self, pile, stations=None, events=None, markers=None, ntracks=12,
            marker_editor_sortable=True, follow=None, controls=True,
            opengl=None, instant_close=False):

        qw.QMainWindow.__init__(self)

        self.instant_close = instant_close

        self.dockwidget_to_toggler = {}
        self.dockwidgets = []

        self.setWindowTitle('Snuffler')

        self.pile_viewer = pile_viewer.PileViewer(
            pile, ntracks_shown_max=ntracks, use_opengl=opengl,
            marker_editor_sortable=marker_editor_sortable,
            panel_parent=self)

        self.marker_editor = self.pile_viewer.marker_editor()
        self.add_panel(
            'Markers', self.marker_editor, visible=False,
            where=qc.Qt.RightDockWidgetArea)
        if stations:
            self.get_view().add_stations(stations)

        if events:
            self.get_view().add_events(events)

            if len(events) == 1:
                self.get_view().set_active_event(events[0])

        if markers:
            self.get_view().add_markers(markers)
            self.get_view().associate_phases_to_events()

        self.tabs = SnufflerTabs(self)
        self.setCentralWidget(self.tabs)
        self.add_tab('Main', self.pile_viewer)

        snufflings = self.pile_viewer.setup_snufflings()
        self.setMenuBar(self.pile_viewer.menu)

        self.main_controls = self.pile_viewer.controls()
        self.add_panel('Main Controls', self.main_controls, visible=controls)
        
        # EOST
        # Activation du widget EOST Hodochrone au démarrage de Snuffler
        # QDockWidget: https://doc.qt.io/qt-5/qdockwidget.html
        eost_hodochrone = None
        for widget in self.get_panel_parent_widget().dockwidgets:
            if "EOST - Hodochrones" in widget.windowTitle():
                eost_hodochrone = widget.findChildren(MyScrollArea)[0]

        try:
           self.add_panel('EOST - Hodochrones', eost_hodochrone, visible=True)
        except Exception as e:
            logger.error(str(e))
            logger.error("Impossible de charger le module : EOST - Hodochrones.")
        # EOST


        self.show()

        # EOST
        try:
            snufflings = self.pile_viewer.get_view().default_snufflings
            for snuffling in snufflings:
                if "EOST - Hodochrones" in snuffling.get_name():
                    snuffling.call()
        except Exception as e:
            logger.error(str(e))
            logger.error("Impossible d'initialiser le module : EOST - Hodochrones.")

        # EOST

        self.get_view().setFocus(qc.Qt.OtherFocusReason)

        sb = self.statusBar()
        sb.clearMessage()
        sb.showMessage('Appuyez sur <?> pour obtenir de l\'aide.')

        snuffler_config = self.pile_viewer.viewer.config

        if snuffler_config.first_start:
            wizard = SnufflerStartWizard(self)

            @qc.pyqtSlot()
            def wizard_finished(result):
                if result == wizard.Accepted:
                    snuffler_config.first_start = False
                    config.write_config(snuffler_config, 'snuffler')

            wizard.finished.connect(wizard_finished)

            wizard.show()

        if follow:
            self.get_view().follow(float(follow))

        self.closing = False

    def sizeHint(self):
        return qc.QSize(1024, 768)
        # return qc.QSize(800, 600) # used for screen shots in tutorial

    def keyPressEvent(self, ev):
        self.get_view().keyPressEvent(ev)

    def get_view(self):
        return self.pile_viewer.get_view()

    def get_panel_parent_widget(self):
        return self

    def add_tab(self, name, widget):
        self.tabs.append_tab(widget, name)

    def remove_tab(self, widget):
        self.tabs.remove_tab(widget)

    def add_panel(self, name, panel, visible=False, volatile=False,
                  where=qc.Qt.BottomDockWidgetArea):

        if not self.dockwidgets:
            self.dockwidgets = []

        dws = [x for x in self.dockwidgets if self.dockWidgetArea(x) == where]

        dockwidget = qw.QDockWidget(name, self)
        self.dockwidgets.append(dockwidget)
        dockwidget.setWidget(panel)
        panel.setParent(dockwidget)
        self.addDockWidget(where, dockwidget)

        if dws:
            self.tabifyDockWidget(dws[-1], dockwidget)

        self.toggle_panel(dockwidget, visible)

        mitem = qw.QAction(name, None)

        def toggle_panel(checked):
            self.toggle_panel(dockwidget, True)

        mitem.triggered.connect(toggle_panel)

        if volatile:
            def visibility(visible):
                if not visible:
                    self.remove_panel(panel)

            dockwidget.visibilityChanged.connect(
                visibility)

        self.get_view().add_panel_toggler(mitem)
        self.dockwidget_to_toggler[dockwidget] = mitem

        if pile_viewer.is_macos:
            tabbars = self.findChildren(qw.QTabBar)
            for tabbar in tabbars:
                tabbar.setShape(qw.QTabBar.TriangularNorth)
                tabbar.setDocumentMode(True)

    def toggle_panel(self, dockwidget, visible):
        if visible is None:
            visible = not dockwidget.isVisible()

        dockwidget.setVisible(visible)
        if visible:
            w = dockwidget.widget()
            try:
                minsize = w.minimumSize()
            except: return None
            w.setMinimumHeight(w.sizeHint().height() + 5)

            def reset_minimum_size():
                import sip
                if not sip.isdeleted(w):
                    w.setMinimumSize(minsize)

            qc.QTimer.singleShot(200, reset_minimum_size)

            dockwidget.setFocus()
            dockwidget.raise_()

    def toggle_marker_editor(self):
        self.toggle_panel(self.marker_editor.parent(), None)

    def toggle_main_controls(self):
        self.toggle_panel(self.main_controls.parent(), None)

    def remove_panel(self, panel):
        dockwidget = panel.parent()
        self.removeDockWidget(dockwidget)
        dockwidget.setParent(None)
        mitem = self.dockwidget_to_toggler[dockwidget]
        self.get_view().remove_panel_toggler(mitem)

    def return_tag(self):
        return self.get_view().return_tag

    def confirm_close(self):
        ret = qw.QMessageBox.question(
            self,
            'Snuffler',
            'Fermer Snuffler ?',
            qw.QMessageBox.Cancel | qw.QMessageBox.Ok,
            qw.QMessageBox.Ok)

        return ret == qw.QMessageBox.Ok

    def closeEvent(self, event):
        if self.instant_close or self.confirm_close():
            self.closing = True
            self.pile_viewer.cleanup()
            event.accept()
        else:
            event.ignore()

    def is_closing(self):
        return self.closing



# https://pyrocko.org - GPLv3
#
# The pyrockoeost Developers, 21st Century
# ---|P------/S----------~Lg----------

from __future__ import absolute_import, print_function, division

import math
import time
import numpy as num

from pyrockoeost.gui.qt_compat import qc, qg, qw

from pyrockoeost.gui.talkie import TalkieConnectionOwner
from pyrockoeost.gui.drum.state import State, TextStyle
from pyrockoeost.gui_util import make_QPolygonF, PhaseMarker
from pyrockoeost import trace, util, pile


def lim(a, x, b):
    return min(max(a, x), b)


class Label(object):
    def __init__(
            self, x, y, label_str,
            anchor='BL',
            style=None,
            keep_inside=None,
            head=None):

        if style is None:
            style = TextStyle()

        text = qg.QTextDocument()
        font = style.qt_font
        if font:
            text.setDefaultFont(font)

        color = style.color.qt_color

        text.setDefaultStyleSheet('span { color: %s; }' % color.name())
        text.setHtml('<span>%s</span>' % label_str)

        self.position = x, y
        self.anchor = anchor
        self.text = text
        self.style = style
        self.keep_inside = keep_inside
        if head:
            self.head = head
        else:
            self.head = (0., 0.)

    def draw(self, p):
        s = self.text.size()
        rect = qc.QRectF(0., 0., s.width(), s.height())

        tx, ty = x, y = self.position
        anchor = self.anchor

        pxs = self.text.defaultFont().pointSize()

        oy = self.head[0] * pxs
        ox = self.head[1]/2. * pxs

        if 'L' in anchor:
            tx -= min(ox*2, rect.width()/2.)

        elif 'R' in anchor:
            tx -= rect.width() - min(ox*2., rect.width()/2.)

        elif 'C' in anchor:
            tx -= rect.width()/2.

        if 'B' in anchor:
            ty -= rect.height() + oy

        elif 'T' in anchor:
            ty += oy

        elif 'M' in anchor:
            ty -= rect.height()/2.

        rect.translate(tx, ty)

        if self.keep_inside:
            keep_inside = self.keep_inside
            if rect.top() < keep_inside.top():
                rect.moveTop(keep_inside.top())

            if rect.bottom() > keep_inside.bottom():
                rect.moveBottom(keep_inside.bottom())

            if rect.left() < keep_inside.left():
                rect.moveLeft(keep_inside.left())

            if rect.right() > keep_inside.right():
                rect.moveRight(keep_inside.right())

        poly = None
        if self.head[0] != 0.:
            l, r, t, b = rect.left(), rect.right(), rect.top(), rect.bottom()

            if 'T' in anchor:
                a, b = t, b
            elif 'B' in anchor:
                a, b = b, t
            elif 'M' in anchor:
                assert False, 'label cannot have head with M alignment'

            c1,  c2 = lim(l, x-ox, r),  lim(l, x+ox, r)

            px = (l, c1, x, c2, r, r, l)
            py = (a, a, y, a, a, b, b)

            poly = make_QPolygonF(px, py)

        tx = rect.left()
        ty = rect.top()

        if self.style.outline or self.style.background_color:
            oldpen = p.pen()
            oldbrush = p.brush()
            if not self.style.outline:
                p.setPen(qg.QPen(qc.Qt.NoPen))

            p.setBrush(self.style.background_color.qt_color)
            if poly:
                p.drawPolygon(poly)
            else:
                p.drawRect(rect)

            if self.style.background_color:
                p.fillRect(rect, self.style.background_color.qt_color)

            p.setPen(oldpen)
            p.setBrush(oldbrush)

        p.translate(tx, ty)
        self.text.drawContents(p)
        p.translate(-tx, -ty)


class ProjectXU(object):
    def __init__(self, xr=(0., 1.), ur=(0., 1.)):
        (self.xmin, self.xmax) = xr
        (self.umin, self.umax) = ur
        self.cxu = (self.umax - self.umin) / (self.xmax - self.xmin)

    def u(self, x):
        return self.umin + (x-self.xmin) * self.cxu

    def x(self, u):
        return self.xmin + (u-self.umin) / self.cxu


class ProjectYV(object):
    def __init__(self, yr=(0., 1.), vr=(0., 1.)):
        (self.ymin, self.ymax) = yr
        (self.vmin, self.vmax) = vr
        self.cyv = (self.vmax - self.vmin) / (self.ymax - self.ymin)

    def v(self, y):
        return self.vmin + (y-self.ymin) * self.cyv

    def y(self, v):
        return self.ymin + (v-self.vmin) / self.cyv


class ProjectXYUV(object):
    def __init__(self, xyr=((0., 1.), (0., 1.)), uvr=((0., 1.), (0., 1.))):
        (self.xmin, self.xmax), (self.ymin, self.ymax) = xyr
        (self.umin, self.umax), (self.vmin, self.vmax) = uvr
        self.cxu = (self.umax - self.umin) / (self.xmax - self.xmin)
        self.cyv = (self.vmax - self.vmin) / (self.ymax - self.ymin)

    def u(self, x):
        return self.umin + (x-self.xmin) * self.cxu

    def v(self, y):
        return self.vmin + (y-self.ymin) * self.cyv

    def x(self, u):
        return self.xmin + (u-self.umin) / self.cxu

    def y(self, v):
        return self.ymin + (v-self.vmin) / self.cyv


class PlotEnv(qg.QPainter):
    def __init__(self, *args):
        qg.QPainter.__init__(self, *args)
        self.umin = 0.
        self.umax = 1.
        self.vmin = 0.
        self.vmax = 1.

    def draw_vline(self, project, x, y0, y1):
        u = project.u(x)
        line = qc.QLineF(u, project.v(y0), u, project.v(y1))
        self.drawLine(line)

    def projector(self, xyr):
        return ProjectXYUV(xyr, self.uvrange)

    @property
    def uvrange(self):
        return (self.umin, self.umax), (self.vmin, self.vmax)


def time_fmt_drumline(t):
    ti = int(t)
    if ti % 60 == 0:
        fmt = '%H:%M'
    else:
        fmt = '%H:%M:%S'

    if ti % (3600*24) == 0:
        fmt = fmt + ' %Y-%m-%d'

    return fmt


class Empty(Exception):
    pass


class DrumLine(qc.QObject, TalkieConnectionOwner):
    def __init__(self, iline, tmin, tmax, traces, state):
        qc.QObject.__init__(self)
        self.traces = traces
        self.tmin = tmin
        self.tmax = tmax
        self.ymin = 0.
        self.ymax = 1.
        self.ymin_data = 0.
        self.ymax_data = 1.
        self.iline = iline
        self._last_mode = None
        self._ydata_cache = {}
        self._time_per_pixel = None
        self.access_counter = 0
        self.talkie_connect(state, 'style.trace_resolution', self._empty_cache)

    def data_range(self, mode='min-max'):
        if not self.traces:
            raise Empty()

        modemap = {
                'min-max': 'minmax',
                'mean-plusminus-1-sigma': 1.,
                'mean-plusminus-2-sigma': 2.,
                'mean-plusminus-4-sigma': 4.,
        }

        if self._last_mode != mode:
            mi, ma = trace.minmax(
                self.traces, key=lambda tr: None, mode=modemap[mode])[None]
            self.ymin_data = mi
            self.ymax_data = ma
            self._last_mode = mode

        return self.ymin_data, self.ymax_data

    def set_yrange(self, ymin, ymax):
        if ymax == ymin:
            ymax = 0.5*(ymin + ymax) + 1.0
            ymin = 0.5*(ymin + ymax) - 1.0

        self.ymin = ymin
        self.ymax = ymax

    @property
    def tyrange(self):
        return (self.tmin, self.tmax), (self.ymin, self.ymax)

    def draw(self, plotenv, markers):
        self._draw_traces(plotenv)
        self._draw_time_label(plotenv)
        self._draw_markers(plotenv, markers)

    def _draw_time_label(self, plotenv):
        text = util.time_to_str(self.tmin, format=time_fmt_drumline(self.tmin))
        font = plotenv.style.label_textstyle.qt_font
        lab = Label(
            font.pointSize(), plotenv.vmin, text,
            anchor='ML', style=plotenv.style.label_textstyle)

        lab.draw(plotenv)

    def _draw_markers(self, plotenv, markers):
        project = plotenv.projector(((self.tmin, self.tmax), (-1., 1.)))
        plotenv.setPen(plotenv.style.marker_color.qt_color)

        rect = qc.QRectF(
            plotenv.umin, 0.,
            plotenv.umax-plotenv.umin, plotenv.widget.height())

        for marker in markers:
            plotenv.draw_vline(project, marker.tmin, -1., 1.)
            s = marker.get_label()
            if s:
                if isinstance(marker, PhaseMarker):
                    anchor = 'BC'
                    v = project.v(-1.)
                else:
                    anchor = 'BR'
                    v = project.v(-1.)

                lab = Label(
                    project.u(marker.tmin), v, s,
                    anchor=anchor,
                    style=plotenv.style.marker_textstyle,
                    keep_inside=rect,
                    head=(1.0, 3.0))

                lab.draw(plotenv)

    def _draw_traces(self, plotenv):
        project = plotenv.projector(self.tyrange)
        tpp = (project.xmax - project.xmin) / (project.umax - project.umin)
        if self._time_per_pixel != tpp:
            self._empty_cache()
            self._time_per_pixel = tpp

        for tr in self.traces:
            udata, vdata = self._projected_trace_data(
                tr,
                project,
                plotenv.style.trace_resolution)

            qpoints = make_QPolygonF(udata, vdata)
            plotenv.setPen(plotenv.style.trace_color.qt_color)
            plotenv.drawPolyline(qpoints)

    def _projected_trace_data(self, tr, project, trace_resolution):
        n = tr.data_len()
        if trace_resolution > 0 \
                and n > 2 \
                and tr.deltat < 0.5 / trace_resolution*self._time_per_pixel:

            spp = int(self._time_per_pixel / tr.deltat / trace_resolution)
            if tr not in self._ydata_cache:
                nok = (tr.data_len() // spp) * spp
                ydata_rs = tr.ydata[:nok].reshape((-1, spp))
                ydata = num.empty((nok // spp)*2)
                ydata[::2] = num.min(ydata_rs, axis=1)
                ydata[1::2] = num.max(ydata_rs, axis=1)
                self._ydata_cache[tr] = ydata
            else:
                ydata = self._ydata_cache[tr]

            udata_min = float(
                project.u(tr.tmin))
            udata_max = float(
                project.u(tr.tmin+0.5*tr.deltat*spp*(ydata.size-1)))
        else:
            ydata = tr.ydata
            udata_min = float(project.u(tr.tmin))
            udata_max = float(project.u(tr.tmin+tr.deltat*(n-1)))

        vdata = project.v(ydata)
        udata = num.linspace(udata_min, udata_max, vdata.size)
        return udata, vdata

    def _empty_cache(self):
        self._ydata_cache = {}


def tlen(x):
    return x.tmax-x.tmin


def is_relevant(x, tmin, tmax):
    return tmax >= x.tmin and x.tmax >= tmin


class MarkerStore(object):
    def __init__(self):
        self.empty()

    def empty(self):
        self._by_tmin = pile.Sorted([], 'tmin')
        self._by_tmax = pile.Sorted([], 'tmax')
        self._by_tlen = pile.Sorted([], tlen)
        self._adjust_minmax()
        self._listeners = []

    def relevant(self, tmin, tmax, selector=None):
        if not self._by_tmin or not is_relevant(self, tmin, tmax):
            return []

        if selector is None:
            return [
                x for x in self._by_tmin.with_key_in(tmin-self.tlenmax, tmax)
                if is_relevant(x, tmin, tmax)]
        else:
            return [
                x for x in self._by_tmin.with_key_in(tmin-self.tlenmax, tmax)
                if is_relevant(x, tmin, tmax) and selector(x)]

    def insert(self, x):
        self._by_tmin.insert(x)
        self._by_tmax.insert(x)
        self._by_tlen.insert(x)

        self._adjust_minmax()
        self._notify_listeners('insert', x)

    def remove(self, x):
        self._by_tmin.remove(x)
        self._by_tmax.remove(x)
        self._by_tlen.remove(x)

        self._adjust_minmax()
        self._notify_listeners('remove', x)

    def insert_many(self, x):
        self._by_tmin.insert_many(x)
        self._by_tmax.insert_many(x)
        self._by_tlen.insert_many(x)

        self._adjust_minmax()
        self._notify_listeners('insert_many', x)

    def remove_many(self, x):
        self._by_tmin.remove_many(x)
        self._by_tmax.remove_many(x)
        self._by_tlen.remove_many(x)

        self._adjust_minmax()
        self._notify_listeners('remove_many', x)

    def add_listener(self, obj):
        self._listeners.append(util.smart_weakref(obj))

    def _adjust_minmax(self):
        if self._by_tmin:
            self.tmin = self._by_tmin.min().tmin
            self.tmax = self._by_tmax.max().tmax
            self.tlenmax = tlen(self._by_tlen.max())
        else:
            self.tmin = None
            self.tmax = None
            self.tlenmax = None

    def _notify_listeners(self, what, x):
        for ref in self._listeners:
            obj = ref()
            if obj:
                obj(what, x)

    def __iter__(self):
        return iter(self._by_tmin)


class DrumViewMain(qw.QWidget, TalkieConnectionOwner):

    def __init__(self, pile, *args):
        qw.QWidget.__init__(self, *args)

        self.setAttribute(qc.Qt.WA_AcceptTouchEvents, True)

        st = self.state = State()
        self.markers = MarkerStore()
        self.markers.add_listener(self._markers_changed)

        self.pile = pile
        self.pile.add_listener(self._pile_changed)

        self._drumlines = {}
        self._wheel_pos = 0
        self._iline_float = None
        self._project_iline_to_screen = ProjectYV(
            (st.iline-0.5, st.iline+st.nlines+0.5), (0., self.height()))
        self._access_counter = 0
        self._waiting_for_first_data = True

        self._init_touch()
        self._init_following()

        self.talkie_connect(
            self.state,
            '',
            self._state_changed,
            drop_args=True)

        self.talkie_connect(
            self.state,
            ['filters', 'nslc', 'tline'],
            self._drop_cached_drumlines,
            drop_args=True)

        self.talkie_connect(
            self.state,
            'style.background_color',
            self._adjust_background_color,
            drop_args=True)

        self.talkie_connect(
            self.state,
            'follow',
            self._adjust_follow,
            drop_args=True)

        self._adjust_background_color()
        self._adjust_follow()

    def goto_data_begin(self):
        if self.pile.tmin:
            self.state.tmin = self.pile.tmin

    def goto_data_end(self):
        if self.pile.tmax:
            self.state.tmax = self.pile.tmax

    def next_nslc(self):
        nslc_ids = sorted(self.pile.nslc_ids.keys())
        if nslc_ids:
            try:
                i = nslc_ids.index(self.state.nslc)
            except ValueError:
                i = -1

            self.state.nslc = nslc_ids[(i+1) % len(nslc_ids)]

    def title(self):
        return ' '.join(x for x in self.state.nslc if x)

    def _state_changed(self):
        self.update()

    def _markers_changed(self, *_):
        self.update()

    def _pile_changed(self, what, content):
        self._adjust_first_data()

        delete = []
        tline = self.state.tline
        for iline in self._drumlines.keys():
            for c in content:
                if c.overlaps(iline*tline, (iline+1)*tline):
                    delete.append(iline)

        for iline in delete:
            del self._drumlines[iline]

        self.update()

    def _init_following(self):
        self._follow_timer = None
        self._following_interrupted = False
        self._following_interrupted_tstart = None

    def interrupt_following(self):
        self._following_interrupted = True
        self._following_interrupted_tstart = time.time()

    def _follow_update(self):
        if self._following_interrupted:
            now = time.time()
            if self._following_interrupted_tstart < now - 20.:
                self._following_interrupted = False
            else:
                return

        now = time.time()
        iline = int(math.ceil(now / self.state.tline))-self.state.nlines
        if iline != self.state.iline:
            self.state.iline = iline

    def _adjust_follow(self):
        follow = self.state.follow
        if follow and not self._follow_timer:
            self._follow_timer = qc.QTimer(self)
            self._follow_timer.timeout.connect(self._follow_update)
            self._follow_timer.setInterval(1000)
            self._follow_update()
            self._follow_timer.start()

        elif not follow and self._follow_timer:
            self._follow_timer.stop()

    def _draw(self, plotenv):
        self._adjust_first_data()
        plotenv.umin = 0.
        plotenv.umax = self.width()
        self._draw_title(plotenv)
        self._draw_time_axis(plotenv)
        self._draw_lines(plotenv)

    def _draw_title(self, plotenv):
        font = plotenv.style.title_textstyle.qt_font
        lab = Label(
            0.5*(plotenv.umin + plotenv.umax),
            font.pointSize(),
            self.title(),
            anchor='TC',
            style=plotenv.style.title_textstyle)

        lab.draw(plotenv)

    def _draw_time_axis(self, plotenv):
        pass

    def _draw_lines(self, plotenv):
        st = self.state
        drumlines_seen = []
        for iline in range(st.iline, st.iline+st.nlines):
            self._update_line(iline)
            drumline = self._drumlines.get(iline, None)
            if drumline:
                drumlines_seen.append(drumline)

        self._autoscale(drumlines_seen)

        top_margin = 50.
        bottom_margin = 50.

        self._project_iline_to_screen = ProjectYV(
                (st.iline-0.5, st.iline+st.nlines-0.5),
                (top_margin, self.height()-bottom_margin))

        for drumline in drumlines_seen:
            plotenv.vmin = self._project_iline_to_screen.v(drumline.iline-0.5)
            plotenv.vmax = self._project_iline_to_screen.v(drumline.iline+0.5)
            markers = self._relevant_markers(drumline.tmin, drumline.tmax)
            drumline.draw(plotenv, markers)
            drumline.access_counter = self._access_counter
            self._access_counter += 1

        drumlines_by_access = sorted(
            self._drumlines.values(), key=lambda dl: dl.access_counter)

        for drumline in drumlines_by_access[:-st.npages_cache*st.nlines]:
            del self._drumlines[drumline.iline]

    def _relevant_markers(self, tmin, tmax):
        return self.markers.relevant(
            tmin, tmax,
            lambda m: not m.nslc_ids or m.match_nslc(self.state.nslc))

    def _autoscale(self, drumlines):
        if not drumlines:
            return

        st = self.state

        data = []
        for drumline in drumlines:
            try:
                data.append(drumline.data_range(st.scaling.base))
            except Empty:
                pass

        if not data:
            data = [[0, 0]]

        mi, ma = num.array(data, dtype=float).T
        gain = st.scaling.gain
        if st.scaling.mode == 'same':
            ymin, ymax = mi.min(), ma.max()
            for drumline in drumlines:
                drumline.set_yrange(ymin/gain, ymax/gain)
        elif st.scaling.mode == 'individual':
            for drumline, ymin, ymax in zip(drumlines, mi, ma):
                drumline.set_yrange(ymin/gain, ymax/gain)
        elif st.scaling.mode == 'fixed':
            for drumline in drumlines:
                drumline.set_yrange(st.scaling.min/gain, st.scaling.max/gain)

    def _update_line(self, iline):

        if iline not in self._drumlines:
            st = self.state
            tmin = iline*st.tline
            tmax = (iline+1)*st.tline
            if st.filters:
                tpad = max(x.tpad() for x in st.filters)
            else:
                tpad = 0.0

            traces = self.pile.all(
                    tmin=iline*st.tline,
                    tmax=(iline+1)*st.tline,
                    tpad=tpad,
                    trace_selector=lambda tr: tr.nslc_id == st.nslc,
                    keep_current_files_open=True,
                    accessor_id=id(self))

            for tr in traces:
                for filter in st.filters:
                    filter.apply(tr)

            self._drumlines[iline] = DrumLine(
                iline, tmin, tmax, traces, self.state)

    def _drop_cached_drumlines(self):
        self._drumlines = {}

    def _adjust_background_color(self):
        color = self.state.style.background_color.qt_color

        p = qg.QPalette()
        p.setColor(qg.QPalette.Background, color)
        self.setAutoFillBackground(True)
        self.setPalette(p)

    def _adjust_first_data(self):
        if self._waiting_for_first_data:
            if self.pile.tmin:
                self.next_nslc()
                self.goto_data_end()
                self._waiting_for_first_data = False

    # qt event handlers

    def paintEvent(self, event):
        plotenv = PlotEnv(self)
        plotenv.style = self.state.style
        plotenv.widget = self

        if plotenv.style.antialiasing:
            plotenv.setRenderHint(qg.QPainter.Antialiasing)

        self._draw(plotenv)

    def event(self, event):

        if event.type() in (
                qc.QEvent.TouchBegin,
                qc.QEvent.TouchUpdate,
                qc.QEvent.TouchEnd,
                qc.QEvent.TouchCancel):

            return self._touch_event(event)

        return qw.QWidget.event(self, event)

    def _init_touch(self):
        self._gesture = None

    def _touch_event(self, event):
        if event.type() == qc.QEvent.TouchBegin:
            self._gesture = DrumGesture(self)
            self._gesture.update(event)

        elif event.type() == qc.QEvent.TouchUpdate:
            self._gesture.update(event)

        elif event.type() == qc.QEvent.TouchEnd:
            self._gesture.update(event)
            self._gesture.end()
            self._gesture = None

        elif event.type() == qc.QEvent.TouchCancel:
            self._gesture.update(event)
            self._gesture.cancel()
            self._gesture = None

        return True

    def wheelEvent(self, event):

        self.interrupt_following()

        self._wheel_pos += event.angleDelta().y()

        n = self._wheel_pos // 120
        self._wheel_pos = self._wheel_pos % 120
        if n == 0:
            return

        amount = max(1., self.state.nlines/24.)
        wdelta = amount * n

        if event.modifiers() & qc.Qt.ControlModifier:
            proj = self._project_iline_to_screen

            anchor = (
                proj.y(event.y()) - proj.ymin) / (proj.ymax - proj.ymin)

            nlines = max(1, self.state.nlines + int(round(wdelta)))

            if self._iline_float is None:
                iline_float = float(self.state.iline)
            else:
                iline_float = self._iline_float

            self._iline_float = iline_float-anchor*wdelta

            self.state.iline = int(round(iline_float))
            self.state.nlines = nlines

        else:
            self.state.iline -= int(wdelta)
            self._iline_float = None

    def keyPressEvent(self, event):

        keytext = str(event.text())

        if event.key() == qc.Qt.Key_PageDown:
            self.interrupt_following()
            self.state.iline += self.state.nlines

        elif event.key() == qc.Qt.Key_PageUp:
            self.interrupt_following()
            self.state.iline -= self.state.nlines

        elif keytext == '+':
            self.state.scaling.gain *= 1.5

        elif keytext == '-':
            self.state.scaling.gain *= 1.0 / 1.5

        elif keytext == ' ':
            self.next_nslc()

        elif keytext == 'p':
            print(self.state)


tline_choices = [10., 30., 60., 120., 300., 600., 1200., 3600.]


class DrumGesture(object):

    def __init__(self, view):
        self._view = view
        self._active = False
        self.begin()

    def begin(self):
        self._active = True
        self._iline = self._view.state.iline
        self._nlines = self._view.state.nlines
        self._tline = self._view.state.tline
        self._proj = self._view._project_iline_to_screen
        self._gain = self._view.state.scaling.gain
        self._max_len_tps = 0

    def end(self):
        self._active = False

    def cancel(self):
        self.end()

    def update(self, event):
        if not self._active:
            return

        tps = event.touchPoints()
        proj = self._proj

        self._max_len_tps = max(len(tps), self._max_len_tps)

        if len(tps) == 1 and self._max_len_tps < 2:
            tp = tps[0]
            iline = proj.y(tp.pos().y())
            iline_start = proj.y(tp.startPos().y())
            idelta = int(round(iline - iline_start))
            iline = self._iline - idelta
            if iline != self._view.state.iline:
                self._view.interrupt_following()
                self._view.state.iline = iline

        if len(tps) == 2:
            # self._view.state.iline = self._iline

            u = [
                (tp.pos().x(), tp.startPos().x())
                for tp in tps]

            y = [
                (proj.y(tp.pos().y()), proj.y(tp.startPos().y()))
                for tp in tps]

            v = [
                (tp.pos().y(), tp.startPos().y())
                for tp in tps]

            if abs(u[0][1] - u[1][1]) < abs(v[0][1] - v[1][1]):

                vclip = self._view.size().height() * 0.05
                d0 = max(vclip, abs(v[0][1] - v[1][1]))
                d1 = max(vclip, abs(v[0][0] - v[1][0]))

                nlines = min(max(1, int(round(self._nlines * (d0 / d1)))), 50)

                idelta = int(round(
                    0.5 * ((y[0][0]+y[1][0]) - (y[0][1]+y[1][1]))
                    * (nlines / self._nlines)))

                iline = self._iline - idelta

                self._view.interrupt_following
                if self._view.state.nlines != nlines \
                        or self._view.state.iline != iline:
                    self._view.interrupt_following()
                    self._view.state.iline = iline
                    self._view.state.nlines = nlines

            else:
                if abs(u[0][0] - u[0][1]) + abs(u[1][0] - u[1][1]) \
                        > abs(v[0][0] - v[0][1]) + abs(v[1][0] - v[1][1]):

                    ustretch = abs(u[0][1] - u[1][1]) \
                        / (abs(u[0][0] - u[1][0])
                           + self._view.size().width() * 0.05)

                    print('ustretch', ustretch)
                    log_tline_choices = num.log(tline_choices)
                    log_tline = num.log(ustretch * self._tline)

                    ichoice = num.argmin(
                        num.abs(log_tline_choices - log_tline))
                    tline = tline_choices[ichoice]
                    # yanchor = 0.5 * (y[0][1] + y[1][1])
                    # r = (yanchor - proj.ymin) / (proj.ymax - proj.ymin)

                    if self._view.state.tline != tline:
                        self._view.state.iline = int(
                            round(self._iline / (tline / self._tline)))
                        self._view.state.tline = tline

                else:
                    vstretch = 10**(
                        -((v[0][0] - v[0][1]) + (v[1][0] - v[1][1]))
                        / self._view.size().height())

                    self._view.state.scaling.gain = self._gain * vstretch
